/*
   File Name:  afile.cc
   Author:     Gary White
   Date:       June 1, 2001

   Description: Extends the file class by adding five methods:

      Archive(<cFileName>, [bSetting])
      Hidden(<cFileName>, [bSetting])
      ReadOnly(<cFileName>, [bSetting])
      System(<cFileName>, [bSetting])

      Release()	// releases the afile object from memory

   The first four methods all work the same to get and/or
   set file attributes.

   <oRef>.Method(<cFileName>, [bSetting])

   oRef        A reference to the afile object.

   cFileName   A string containing the file name to get/set
               the attribute.

   bSetting    An optional boolean parameter.  If omitted, the 
               existing value of the setting is returned.  If 
               bSetting is supplied, the attribute specified 
               by the method name is set to the value supplied
               and the method returns TRUE, or FALSE indicating
               the success, or failure to set the value.

   Example:

      // is the file filename.txt to readonly?
      set procedure to afile.cc additive
      f = new afile()
      bReadOnly = f.ReadOnly('filename.txt')

      // set the file filename.txt to read-only
      set procedure to afile.cc additive
      f = new afile()
      f.ReadOnly('filename.txt', true)

      // set the file filename.txt to read-write
      set procedure to afile.cc additive
      f = new afile()
      f.ReadOnly('filename.txt', false)


   Note: If a method fails to get the file attributes,
         GetLastError does not work and all I can do 
         is report a general failure message.  Except 
         for that limitation, this seems to work
*/

#define FILE_ATTRIBUTE_READONLY        0x00000001
#define FILE_ATTRIBUTE_HIDDEN          0x00000002
#define FILE_ATTRIBUTE_SYSTEM          0x00000004
#define FILE_ATTRIBUTE_ARCHIVE         0x00000020

class afile of file
   protect init, SetAttribute, ClearAttribute

   this.Init()

   function Init
      // prototype a couple of API calls
      if type('GetFileAttributes') # 'FP'
         extern culong GetFileAttributes(cstring) kernel32 ;
            from 'GetFileAttributesA'
      endif
      if type('SetFileAttributes') # 'FP'
         extern clogical SetFileAttributes(cstring, cint) kernel32 ;
            from 'SetFileAttributesA'
      endif
   return

   function Archive(cName, bArchive)
      local n, bRet, nParams
      bRet = false
      n = 0
      nParams = argcount()
      if nParams > 0
         if type('argvector(1)') == 'C'
            if nParams = 1
               n := GetFileAttributes(cName)
               bRet = n # 0xFFFFFFFF
               if bRet
                  bRet := bitand(n, FILE_ATTRIBUTE_ARCHIVE) = ;
                     FILE_ATTRIBUTE_ARCHIVE
               endif
            else
               if type('argvector(2)') = 'L'
                  if bArchive
                     bRet := this.SetAttribute(cName, FILE_ATTRIBUTE_ARCHIVE)
                  else
                     bRet := this.ClearAttribute(cName, FILE_ATTRIBUTE_ARCHIVE)
                  endif
               else
                  throw new afError(163, ;
                     'Error:  Data type mismatch.  Expecting logical',;
                     program(), 177)
               endif
            endif
         else
            throw new afError(163, ;
               'Error:  Data type mismatch.  Expecting character',;
               program(), 177)
         endif
      endif
   return bRet

   function Hidden(cName, bHidden)
      local n, bRet, nParams
      bRet = false
      n = 0
      nParams = argcount()
      if nParams > 0
         if type('argvector(1)') == 'C'
            if nParams = 1
               n := GetFileAttributes(cName)
               if n = 0xFFFFFFFF	// unsuccessful
                  throw new afError(0, 'Error getting file attributes', ;
                     program(), 114)
               endif
               bRet := bitand(n, FILE_ATTRIBUTE_HIDDEN) = ;
                  FILE_ATTRIBUTE_HIDDEN
            else
               if type('argvector(2)') = 'L'
                  if bHidden
                     bRet := this.SetAttribute(cName, FILE_ATTRIBUTE_HIDDEN)
                  else
                     bRet := this.ClearAttribute(cName, FILE_ATTRIBUTE_HIDDEN)
                  endif
               else
                  throw new afError(163, ;
                     'Error:  Data type mismatch.  Expecting logical',;
                     program(), 106)
               endif
            endif
         else
            throw new afError(163, ;
               'Error:  Data type mismatch.  Expecting character',;
               program(), 106)
         endif
      endif
   return bRet

   function ReadOnly(cName, bReadOnly)
      local n, bRet, nParams
      bRet = false
      n = 0
      nParams = argcount()
      if nParams > 0
         if type('argvector(1)') == 'C'
            if nParams = 1
               n := GetFileAttributes(cName)
               if n = 0xFFFFFFFF	// unsuccessful
                  throw new afError(0, 'Error getting file attributes', ;
                     program(), 77)
               endif
               bRet := bitand(n, FILE_ATTRIBUTE_READONLY) = ;
                  FILE_ATTRIBUTE_READONLY
            else
               if type('argvector(2)') = 'L'
                  if bReadOnly
                     bRet := this.SetAttribute(cName, FILE_ATTRIBUTE_READONLY)
                  else
                     bRet := this.ClearAttribute(cName, FILE_ATTRIBUTE_READONLY)
                  endif
               else
                  throw new afError(163, ;
                     'Error:  Data type mismatch.  Expecting logical',;
                     program(), 70)
               endif
            endif
         else
            throw new afError(163, ;
               'Error:  Data type mismatch.  Expecting character',;
               program(), 70)
         endif
      endif
   return bRet

   function System(cName, bSystem)
      local n, bRet, nParams
      bRet = false
      n = 0
      nParams = argcount()
      if nParams > 0
         if type('argvector(1)') == 'C'
            if nParams = 1
               n := GetFileAttributes(cName)
               bRet = n # 0xFFFFFFFF
               if bRet
                  bRet := bitand(n, FILE_ATTRIBUTE_SYSTEM) = ;
                     FILE_ATTRIBUTE_SYSTEM
               endif
            else
               if type('argvector(2)') = 'L'
                  if bSystem
                     bRet := this.SetAttribute(cName, FILE_ATTRIBUTE_SYSTEM)
                  else
                     bRet := this.ClearAttribute(cName, FILE_ATTRIBUTE_SYSTEM)
                  endif
               else
                  throw new afError(163, ;
                     'Error:  Data type mismatch.  Expecting logical',;
                     program(), 142)
               endif
            endif
         else
            throw new afError(163, ;
               'Error:  Data type mismatch.  Expecting character',;
               program(), 142)
         endif
      endif
   return bRet

   function SetAttribute(cFile, nAttrib)
   	local bRet, n
      bRet = true
      n = GetFileAttributes(cFile)
      if bitand(n, nAttrib) = 0
      	bRet := SetFileAttributes(cFile, bitor(n, nAttrib))
      endif
   return bRet

   function ClearAttribute(cFile, nAttrib)
      local bRet
      bRet = true
      n = GetFileAttributes(cFile)
      if bitand(n, nAttrib) = nAttrib
         bRet := SetFileAttributes(cFile, bitand(n, bitnot(nAttrib)))
      endif
   return bRet

   function Release
      this.close()
      release object this
   return

endclass

class afError(nCode, cMsg, cRoutine, nLine) of exception
   this.code := nCode
   this.message := cMsg
   this.filename := cRoutine
   this.lineno := nLine
endclass

