/*
   Class:   ColorGetter

   Author:  Daniel Howard (sproket@total.net)

   History: 
      1.0   December 12, 1998
            Initial Release

   Inherits: dContainer

   Dependants: dControls.cc, ini.cc, ini2.cc

   Properties:

   Methods:
      Init        -  Extends the dContainer's Init. If you override this
                     make sure to call Super::Init().
                     This method creates the colorbutton at the right
                     edge of the entryfield. This is so you can widen
                     the entryfield to whatever width you want and the 
                     button will always appear in the right place.
                     Works with Char metric only.
                     It also defines all the color names found in the
                     ini file.

   Events:
      ColorButton_OnClick - 
                     Calls GetColor() and assigns the color chosen to
                     the object unless the user clicks cancel.

      ColorField_OnChange -
                     Sets the color entered if it's valid.

      Setter_OnChange -
                     Sets the color of the colorbutton to the selected
                     color. 

      Setter_CanChange - 
                     Allows only valid color strings either Hex or 
                     color names.

      Setter_BeforGetValue -
                     Morphs the color from a hex code into a more user
                     friendly color name if one exists. Creates an INI
                     file called ColorGetter.ini which you can add 
                     new color names to.
                     
      
   Notes:
      I use Jay Parsons' RGBToHex routine as well as a HexToRGB
      routine I wrote. They are in the class CG_Converter at the 
      bottom of the file. 

      There's a small bug with colors. BtnText/black seems to not
      change the color property. Tried this in the inspector
      with the same results. It works if BtnText is changed to 
      "white".  

      I use a slighly modified ini and an extended ini called ini2.
      Ini2 simply calls the SectionArrayParse2arrays() method for
      you when you use GetSection(). Read the ini2.cc comments for
      more details.

*----------------------------------------------------------------------
This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug reports/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/

class COLORGETTER(parentObj) of DCONTAINER(parentObj) custom
   with (this)
      left = 0.8333
      top = 4.2
      width = 34.5
      height = 2.3
      metric = 0	// Chars
   endwith


   this.COLORRECTANGLE = new CG_RECTANGLE(this)
   with (this.COLORRECTANGLE)
      left = 0.1667
      top = -0.05
      width = 33.1667
      height = 2
      text = "Color"
      pageno = 0
   endwith


   this.COLORFIELD = new CG_ENTRYFIELD(this)
   with (this.COLORFIELD)
      OnChange = Class::ColorField_onChange
      height = 1
      left = 1.1667
      top = 0.65
      width = 31
      value = ""
      speedtip = "Enter a color"
   endwith

   this.Setter.Value := space(25)
   this.Setter.CanChange := Class::Setter_CanChange
   this.Setter.OnChange := Class::Setter_onChange
   this.Setter.BeforeGetValue := Class::Setter_BeforeGetValue

   Function Init
      Super::Init()
      this.COLORBUTTON = new CG_PushButton(this)
      with (this.COLORBUTTON)
         onClick = class::ColorButton_ONCLICK
         height = this.ColorField.Height - 0.15
         left = this.ColorField.Width - 2.17
         top = 0.75
         width = 3
         value = false
         speedtip = "Click to select a color"
      endwith

      // Define all color names in the ini file so we won't crash
      // on color names
      local a
      a = {"0x0000FF=red","0xFF0000=blue","0x008000=green","0x00FFFF=yellow","0x0080FF=orange","0x800080=purple","0x808080=gray","0xFF80FF=pink","0x000000=black","0x004080=brown","0xFFFFFF=white"}

      private oldPath
      oldPath = set("PATH")
      set path to
      if not file("ColorGet.ini")
         local f
         f = New File()
         f.Create("ColorGet.ini")
         f.WriteLn("[Colors]")
         local j
         for j = 1 to a.Size
            f.WriteLn(a[j])
         next j
         f.Close()
      endif
      set path to &oldPath.

      Set procedure to ini2.cc additive
      local i,j
      i = New Ini2("ColorGet.ini")
      i.GetSection("Colors")
      private m_macro
      for j = 1 to i.EntriesArray.Size
         m_macro = "define color " + i.ValuesArray[j] + " " + CG_Converter::HexToRGB(i.EntriesArray[j])
         &m_macro.
      next j         
   return

   function ColorButton_onClick
      local c
      c = GetColor("ColorGetter")

      if not empty(c)
         this.Parent.Setter.Value := CG_Converter::RgbToHex(c)
      endif                              
   return

   Function Setter_OnChange
      local n
      n = at("/",this.Owner.ColorButton.colorNormal)
      local c
      c = left(this.Owner.ColorButton.ColorNormal,n) + this.Value
      this.Owner.ColorButton.ColorNormal := c
      this.Owner.ColorField.Value := this.Value
   return

   // Check to see if it's a basic color. If it is use the color name
   // which is more user friendly than a Hex number. Create an ini file
   // and store it so the user can add to the basic colors. 
   Function Setter_BeforeGetValue
      local i
      Set procedure to ini.cc additive
      i = New Ini("ColorGet.ini")
      local c
      c = i.GetValue("Colors",this.Value)
      if not empty(c)
         return c 
      endif
   return this.Value

   Function ColorField_OnChange
      this.Parent.Setter.Value := this.Value
      this.Value := this.Parent.Setter.Value
   return

   Function Setter_CanChange
   para xValue

      local lValid
      lValid = true

      local f
      f = New form()
      try
         f.ColorNormal = xValue
      catch (exception e)
         lValid := false
      endtry

      if not lValid
         msgbox("Invalid color!","Color Getter")
      endif
   return ( lValid )
endclass

Class CG_EntryField(parentOBJ) of dEntryField(parentOBJ)
EndClass

Class CG_Rectangle(parentOBJ) of dRectangle(parentOBJ)
EndClass

Class CG_PushButton(parentOBJ) of dPushButton(parentOBJ)
   with (this)
      height = 1
      left = 10.1667
      top = 4.5
      width = 3
      text = ""
      metric = 0	// Chars
      speedBar = true
      upBitmap = "RESOURCE #144"
   endwith
EndClass


Class CG_Converter
   FUNCTION RGBToHex(RGBString)
   local nComma, nVal, cStr, cRet
      cStr = RGBString
      *-- Red
      nComma = at(",", cStr)
      nVal = val(left(cStr, nComma-1))
      cStr = substr(cStr, nComma+1)
      cRet = itoh(nVal,2)

      *-- Green
      nComma = at(",", cStr)
      nVal = val(left(cStr, nComma-1))
      cStr = substr(cStr, nComma+1)
      cRet = itoh(nVal,2)+cRet
      *-- Blue
   RETURN "0x"+itoh(val(cStr),2)+cRet

   FUNCTION HexToRGB(HexString)
      local cRet
      cRet = ""

   *   0x112233
      // Red
      cRet += htoi(substr(HexString,7,2)) + ","

      // Green
      cRet += htoi(substr(HexString,5,2)) + ","

      // Blue
      cRet += htoi(substr(HexString,3,2))
   RETURN cRet
EndClass

