Function crossProd
/*
   -----------------------------------------------------------------------
   Programmer..: Marko Mihorko, Roland Wingerter
   Date........: 08 Mar 2003
   Notes.......: Calculate cross product of n arrays.
					  Cross product (aka cartesian product) is defined as
					  the set of all ordered pairs <a, b>, where 
                 a is an element of A and b is an element of B, 
   Written for.: dBASE PLUS
   Rev. History: 11/12/2002 0.1 - Original version posted by Marko Mihorko 
                                  in NG dbase.programming
                 02/06/2003 0.2 - Added header; included in dUFLP 
                 24 Feb 2003 0.3 - RW fixed bug in usage example,
                                  simplified code, added comments. 
                                  Changed filename to crossProduct.prg
                                  (was: crossProd.prg). 
                 07 Mar 2003 0.4 - Code revised by RW. 
                                  Changed to work with different parameters: 
                                  call with one array, n arrays, array of arrays. 
                                  Changed variable names. 
                 08 Mar 2003 0.5 - Fixed bug introduced in ver 0.4. Changed 
                                  variables n, i to private in crossBuild. 
   Called by...: Any
   Calls.......: crossBuild(), isFlat()
   Usage.......: crossProd(array), where
   				      array can have 1..n arrays,
                     each array can have 1..n elements.
                 or
                 crossProd(arrayOfArrays)

   Example.....: aResult = crossProd({1,2},{'a','b'})
   				  aResult will hold an array of four arrays:
                 {{1,'a'},{1,'b'},{2,'a'},{2,'b'}}
					  aResult = crossProd({1,2,3},{'a','b'},{5,6}) 
                 aResult = crossProd({1}) 
   Returns.....: Array of arrays  
   Parameters..: one array, n Arrays or arrayofArrays
   ---------------------------------------------------------------------
*/

   local aResult, i, nArgCount, oArray 

   /*
   ---------------------------------------------------------------------------
   Parameter checking: check arguments passed and act accordingly.
   The following arguments are accepted:
   - 1 array
   - n arrays 
   - array of 1..n arrays
   ---------------------------------------------------------------------------
   */

   nArgCount = argCount()

   do case

      // nArgCount = 1
   	case nArgCount = 1 and type("argVector(1)") = 'A' ;
         and argVector(1).Size = 1 ;
         and type("argVector(1)[1]") = 'A' 

         // We have an array containing one array -
         // we are able to deal with that. 
         oArray = argVector(1)[1]
         return crossBuild( oArray, oArray )

      case nArgCount = 1 and type("argVector(1)") = 'A' ;
         and argVector(1).size > 1 ;
         and type("argVector(1)[1]") = 'A'

         // Make sure we have an array of arrays, but not an array 
         // which contains arrays and other elements 
         if not isFlat( argVector(1) ) 
         	oArray = argVector(1)
         else
            msgbox("Cannot process arrays mixed with other elements!",;
                   "Error", 16)
            return -1
         endif

      case nArgCount = 1 and type("argVector(1)") = 'A' ;
         and type("argVector(1)[1]") # 'A'

         // Make sure we have an array of simple elements, and only
         // simple elements, no nested arrays. 
         if isFlat ( argVector(1) )
            oArray = argVector(1)
            return crossBuild( oArray, oArray )
   		else
            msgbox("Cannot process arrays mixed with other elements!",;
                   "Error", 16)
            return -1
         endif

      // nArgCount > 1
      case nArgCount > 1 and not isFlat( {argVector(1)} )
         // We have n arrays ...
         oArray = new array()
         for i=1 to nArgCount
            oArray.Add(argVector(i))
         next i
      
      otherwise
         // Just in case we missed something ...
         msgbox("This function requires 1..n arrays as parameter!" ;
               + chr(10)+chr(13)+"Please read instructions!", ; 
               "Error calling crossProduct",16) 
         return-1
   endcase

   /* 
   ---------------------------------------------------------------------------
   If we arrive here, we have 1..n arrays to process. 
   The processing is done by recursive calls to crossBuild. 
   The first time, crossBuild is called with array1 and array2. 
   After that, it is called with the current result array x and the next array 
   until all arrays have been processed. 
   ---------------------------------------------------------------------------
   */

   // Assign first array as start value 
   aResult = oArray[1]

   for i = 2 to oArray.size
      // Build crossproduct from current aResult and next array 
      aResult = crossBuild( aResult, oArray[i] )
   endfor

return aResult
// EoF: crossProduct()

Function crossBuild
/*
   -----------------------------------------------------------------------
   Programmer..: Marko Mihorko, Roland Wingerter
   Date........: 27 Feb 2003
   Notes.......: Builds cross product of two arrays.  
                 Auxiliary function for crossProduct(), can also 
                 be used on its own. 
                 Use crossProduct to process an array of arrays 
                 (or a simple array with one or more elements). 

   Written for.: dBASE PLUS
   Rev. History: 12 Nov 2002 - Original published by Marko Mihorko 
                                  in NG dbase.programming
                 06 Feb 2003 - Added header 
                 24 Feb 2003 - Two bugfixes by Roland Wingerter.
                                 Bugfix 1: avoid empty elements
                                 (logical value false) in result. 
                                 Bugfix 2: avoid converting array elements to 
                                 type string by using acopy. 
                                 Simplified code, changed variables to local, 
                                 added comments. 
                 27 Feb 2003 - Updated header
                   
   Usage.......: crossBuild(array1, array2)
   Example.....: aResult = crossBuild({1,2},{'a','b'})
   				  aResult will hold an array of four arrays:
                 {{1,'a'},{1,'b'},{2,'a'},{2,'b'}}
   Called by...: crossProduct()
   ---------------------------------------------------------------------
   */
      parameters  arr1,arr2

      local aResult, aTemp 

      aResult = new array()
  
      // For each element in array1
      for i=1 to arr1.size  
         // And each element in array2
         // Join elements from array1 and array2
         for n=1 to arr2.Size
            if type("arr1[i]") = 'A'
               // Create copy of arr1[i] 
               aTemp = new array(arr1[i].Size)
               acopy(arr1[i],aTemp)
               // Add one element from arr2
               aTemp.Add(arr2[n])
               // Add temp array to result
               aResult.Add(aTemp)
            else
               aResult.Add({arr1[i],arr2[n]}) 
            endif
      endfor
     endfor

   return aResult
   // EoP: crossBuild()


   FUNCTION isFlat ( oArray ) 
   /*
     -----------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 07 Mar 2003
     Notes.......: Returns true if array does not contain other arrays
     Written for.: dBASE PLUS
     Rev. History: None
     ---------------------------------------------------------------------
   */
      local lRetVal, n
      private xValue

      lRetVal = true
      for n=1 to oArray.size
         xValue = oArray[n]
         if type("xValue") = 'A'
            lRetVal = false
            exit
         endif
      endfor

   Return lRetVal
   // EoP: isFlat()


