// -- File: MATHEX.CC

CLASS MathEx of Math
/*
   ----------------------------------------------------------------------
   Description:

      This class contains many useful math, trigonometric, 
      statistical and random number routines.

   Programmers: 

      Ken Mayer
      Jay Parsons
      Ron Allen
      Oktay Amiry
      Tony Lima
      Michael Nuwer

   History: 

      11/25/1996 -- original version of these routines as a class.
      September 13, 2004 -- Added ExcelFV() and ExcelPV() by
                            Michael Nuwer

   Instantiation: 

      set procedure to :dUFLP:mathex.cc additive
      _app.Math = new MathEx()

    **** NOTE: If you do not use the release() method of MathEx,
         you will not be able to release the references to the
         Root and StringEx classes that are created in the 
         init() method.

   Methods:
      More details on each of these is contained in the commentary for
      each method ...
   
      Init()
        Initializes classes necessary

      Release()
        Releases this object (and others) from memory

      ACosec()
        Inverse CoSecant

      ACosH()
        Inverse hyperbolic cosine

      ACot()
        Inverse CoTangent

      ACoTH()
        Inverse hyperbolic CoTangent

      ACsCH()
        Inverse hyperbolic CoSecant
 
      AHav()
        Inverse haversine

      ASec()
        Inverse Secant

      ASecH()
        Inverse hyperbolic Secant

      ASinH()
        Inverse hyperbolic sine

      ATanH()
        Inverse hyperbolic tangent

      Combos()
        Combinations treat "1, 3" as the same as "3, 1", unlike 
        permutations.  This gives the games needed for a round robin
        and helps with figuring odds of most state lotteries.

      ContRate()
        Rate if compounding is continuous

      CoSecant()
        CoSecant of an angle

      CosH()
        Hyperbolic cosine

      CoTan()
        CoTangent of an angle
 
      CoTH()
        Hyperbolic cotangent

      CsCH()
        Hyperbolic coSecant

      Dice()
        Returns a random integer between 1 and maximum supplied

      Discount()
        Compute present value of amount to be received at end of n 
        periods, given rate

      ExcelFV() 
      ExcelPV()
        Excel versions of FV() and PV() ...

      Factorial()
        Returns factorial of a number

      FuturVal()
        Computes future value of amount at compound interest

      FVIrr()
        Returns same roots as irr2() but faster using net future value

      FVXIrr()
        Combines FVirr() and Xirr(), net future value with random 
        dates

      GCD()
        Greatest common divisor

      Hav()
        Haversine of an angle in radians

      isPrime()
        Returns true if number is prime

      Irr2()
        Returns internal rate of return

      MIrr()
        Modified IRR

      MultDice()
        Rolls multiple dice of specified number of sides, returns 
        total

      NPV2()
        Net Present Value

      ParseDice()
        Parses out and calculates dice based on a 'standard' dice 
        string (RPG format)

      Permutes()
        Permutations of n items taken x at a time

      RandSel()
        Random selection of integers

      Rate()
        Rate of periodic interest needed to produce future value from
        present value

      RMS()
        Root-mean-square

      SampleVar()
        Sample Variance

      SD()
        Standard Deviation

      Secant()
        Secant of an angle

      SecH()
        Hyperbolic secant of an angle

      SinH()
        Hyperbolic sine of an angle

      StnArea()
        Area of standard normal distribution function between mean 
        and x std

      StnDevs()
        Calculates Z standard deviations

      StnDiff()
        Tests confidence that sample is significantly unlike expected
        mean

      StnY()
        Value of standard normal dist function

      StnZ()
        Lookup table to find values of Z

      TanH()
        Hyperbolic tangent of an angle

      XIrr()
        Calculate internal rate of return from cashflows on random 
        dates

      XMIrr()
        Calculate modified internal rate of return from cashflows on
        random dates

      ZeroIn()
        Finds a zero of a continuous function

   ** Note that this cc file uses Source Aliasing, specifically the
   ** dUFLP source Alias. If you do not know how to set up
   ** a Source Alias, see the instructions in WHATS.NEW at
   ** the top of the file, or in README.TXT.
   ** Any files referenced by :dUFLP:filename  must be
   ** included in an executable if you are building 
   ** one ... make sure that your project includes these
   ** files.
   **
   ** This particular program uses:
   **           stringex.cc
   **           root.cc
   **           setproc.prg

   ---------------------------------------------------------------------- 

*/

   CLASS::Init()

   FUNCTION Init
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 11/10/1996
     Notes.......: Seeds random number table and sets up string and
                   root objects
     Written for.: Visual dBASE 5.5
     Rev. History: 11/10/1996 -- Original Code
     Calls.......: string.cc
     Usage.......: MathEx.Init()
     Example.....: _app.MathEx.Init()
     Returns.....: Nul
     Parameters..: None
    ------------------------------------------------------------------- */
   
      PROTECT StringProc, RootProc
      rand(-1)
      set procedure to :dUFLP:SetProc.prg additive
      if .not. SetProc( ":dUFLP:STRINGEX.CC" )
         set procedure to :dUFLP:stringex.cc additive
         this.StringProc = true
      else
         this.StringProc = false
      endif
      this.String        = new stringex()
      this.String.parent = this

      if .not. SetProc( ":dUFLP:ROOT.CC" )
         set procedure to :dUFLP:root.cc additive
         this.RootProc = true
      else
         this.RootProc = false
      endif
      this.Root        = new root()
      this.Root.parent = this
      
   RETURN NULL

   FUNCTION Release
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 11/24/1996
     Notes.......: Used to release this object from memory.
     Written for.: Visual dBASE 5.5
     Rev. History: 11/24/1996 -- Original Code
     Calls.......: none
     Usage.......: mathex.Release()
     Example.....: _app.mathex.Release()
                   _app.mathex = NULL   && not required, but a good 
                                        && idea
     Returns.....: Null
     Parameters..: None
    ------------------------------------------------------------------- */

      this.String.release()
      this.String = null
      this.Root.release()
      this.Root = null
      if this.StringProc
         close procedure :dUFLP:stringex.cc
      endif
      if this.RootProc
         close procedure :dUFLP:root.cc
      endif
      release object this

   RETURN NULL
   // -- EoM: Release()

   FUNCTION ACosec(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse CoSecant - angle size in radians for a 
                   given CoSecant value n
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ACosec(<n>)
     Example.....: ? _app.Math.ACosec(48)
     Returns.....: Numeric
     Parameters..: n = CoSecant value
    ------------------------------------------------------------------- */
   
   RETURN ( this.asin( 1 / (n) ) )
   // -- EoM: ACosec

   FUNCTION ACosH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Inverse hyperbolic cosine - angle size in radians
                   for given hyperbolic cosine X
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ACosH(<n>)
     Example.....: ? _app.Math.ACosH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic cosine
    ------------------------------------------------------------------- */
      
   RETURN ( this.log((n) + this.sqrt((n)*(n)-1) ) )
   // -- EoM: ACosH

   FUNCTION ACot(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse CoTangent - angle size in radians for a 
                   given CoTangent value n
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ACot(<n>)
     Example.....: ? _app.Math.ACot(48)
     Returns.....: Numeric
     Parameters..: n = CoTangent value
    ------------------------------------------------------------------- */

   RETURN ( this.atan( 1 / (n) ) )
   // -- EoM: ACot()

   FUNCTION ACoTH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/06/1996
     Notes.......: Inverse hyperbolic CoTangent - angle size in 
                   radians for given hyperbolic CoTangent
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ACoTH(<n>)
     Example.....: ? _app.Math.ACoTH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic CoTangent
    ------------------------------------------------------------------- */

   RETURN ( this.log( ((n) + 1) / ((n) - 1) ) / 2 )
   // -- EoM: ACoTH()

   FUNCTION ACsCH(n)
   /* -----------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse hyperbolic CoSecant - angle size in radians
                   for given hyperbolic CoSecant
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ACsCH(<nX>)
     Example.....: ? oMath.ACsCH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic CoSecant
    ----------------------------------------------------------------------- */

   RETURN ( this.log( (sign(n) * this.sqrt((n)*(n) + 1) + 1 )/(n) ) )
   // -- EoM: ACsCH()

   FUNCTION AHav(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse haversine - angle size in radians for 
                   given haversine
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.AHav(<n>)
     Example.....: ? _app.Math.AHav( .5 )
     Returns.....: Numeric
     Parameters..: n = haversine value
    ------------------------------------------------------------------- */

   RETURN ( this.acos( 1 - 2 * (n) ) )
   // -- EoM: AHav()

   FUNCTION ASec(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse Secant - angle size in radians for a given
                   Secant value n
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ASec(<n>)
     Example.....: ? _app.Math.ASec(48)
     Returns.....: Numeric
     Parameters..: n = Secant value
    ------------------------------------------------------------------- */

   RETURN ( this.acos( 1 /(n) ) )
   // -- EoM: ASec()

   FUNCTION ASecH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse hyperbolic Secant - angle size in radians
                   for given hyperbolic Secant
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows. 
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ASecH(<n>)
     Example.....: ? _app.Math.ASecH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic Secant
    ------------------------------------------------------------------- */

   RETURN ( this.log( ( this.sqrt(1 - (n)*(n)) + 1 ) / (n) ) )
   // -- EoM: ASecH()

   FUNCTION ASinH( n )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Inverse hyperbolic sine - angle size in radians
                   for given hyperbolic sine nX
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ASinH(<nX>)
     Example.....: ? oMath.ASinH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic sine
    ------------------------------------------------------------------- */

   RETURN ( this.log( (n) + this.sqrt((n)*(n)+1) ) )
   // -- EoM: ASinH()
   
   FUNCTION ATanH
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Inverse hyperbolic tangent - angle size in radians
                   for given hyperbolic tangent
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.ATanH(<n>)
     Example.....: ? _app.Math.ATanH(48)
     Returns.....: Numeric
     Parameters..: n = hyperbolic tangent
    ------------------------------------------------------------------- */

   RETURN ( this.log( (1 + (n)) / (1 - (n)) ) / 2 )
   // -- EoM: ATanH()

   FUNCTION Combos( nNum, nHowMany )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 03/01/1992
     Notes.......: Combinations, similar to Permutations
                   Combinations treat "1, 3" as the same as
                   "3, 1", unlike permutations.  This gives the
                   games needed for a round robin and helps with
                   figuring odds of most state lotteries.
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   12/05/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: Combos(<nNum>,<nHowMany>)
     Example.....: ?Combos(10,2)
     Returns.....: Numeric
     Parameters..: nNum     = number of items in the entire set
                   nHowMany = number to be used at once
    ------------------------------------------------------------------- */
      
      local nResult, nCounter
      store 1 to nResult
      for nCounter = 1 to nHowmany
          nResult = nResult * ( nNum + 1 - nCounter ) / nCounter
      endfor
      
   RETURN nResult
   // -- Combos()

   FUNCTION ContRate( nFutVal, nPresVal, nYears )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Rate if compounding is continuous.  Periods must 
                   be years.
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.Rate()
     Usage.......: math.ContRate(<nFutVal>,<nPresVal>,<nYears>)
     Example.....: ? _app.Math.ContRate( 50000, 10000, 4.5 )
     Returns.....: Numeric
     Parameters..: nFutVal  = Future Value
                   nPresVal = Present Value
                   nYears   = Number of years to calculate for
    ------------------------------------------------------------------- */

   RETURN this.log ( 1 + CLASS::Rate( (nFutVal), (nPresVal), (nYears) ) )
   // -- EoM: ContRate()

   FUNCTION CoSecant(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: CoSecant of an angle n in radians
                   Indeterminate and returns error if n = 0 or 180 
                   deg., i. e. mod( abs(n), pi() ) = 0
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.CoSecant(<n>)
     Example.....: ? _app.Math.CoSecant(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( 1 / this.sin(n) )
   // -- EoM: CoSecant()

   FUNCTION CosH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Hyperbolic cosine of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.CosH(<nX>)
     Example.....: ? _app.Math.CosH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( (this.exp(n) + this.exp(-(n) )) / 2 )
   // -- EoM: CosH()

   FUNCTION CoTan(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: CoTangent of an angle n in radians. Indeterminate 
                   and returns error if n = 0 or 180 deg.,
                       i. e. mod( abs(n), pi() ) = 0
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.CoTan(<n>)
     Example.....: ? _app.Math.CoTan(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( 1 / this.tan(n) )
   // -- EoM: CoTan()

   FUNCTION CoTH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Hyperbolic cotangent of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.SinH()
                   math.CosH()
     Usage.......: math.CoTH(<n>)
     Example.....: ? _app.Math.CoTH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( CLASS::CosH(n) / CLASS::SinH(n) )
   // -- EoM: CoTH()

   FUNCTION CsCH( n )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Hyperbolic coSecant of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.SinH()
     Usage.......: math.CsCH(<n>)
     Example.....: ? _app.Math.CsCH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( 1 / CLASS::SinH(n) )
   // -- EoM: CsCH()

   FUNCTION Dice( nSides ) 
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 11/10/1996
     Notes.......: A small function used to determine a random 
                   integer from 1 to x. Used for gaming and 
                   simulation purposes. 
     Written for.: dBASE IV, 1.1 
     Rev. History: 05/23/1991 - original function. 
                   Fall, 1994 -- revised for dBASE 5.0 for Windows. 
                                 Jay Parsons, 72662,1302 
                   11/10/1996 -- Modified to work with this class.
     Calls.......: None 
     Usage.......: mathex.Dice(<nSides>) 
     Example.....: nVal = _app.MathEx.Dice(4)  
     Returns.....: Random # between 1 and <nSides> 
     Parameters..: nSides = # of sides of die to be cast ... 
                            (RPG dice include 4, 6 (standard), 8, 10,
                            12, 20, 100  
    ------------------------------------------------------------------- */
 
   RETURN this.int(this.random() * (nSides)) + 1 
   // -- EoM: Dice()

   FUNCTION Discount( nFuturVal, nRate, nPeriods )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Compute the present value of an amount to be 
                   received at the end of a number of periods given 
                   a periodic interest rate.
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302                                     ,
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None                         
     Usage.......: math.Discount(<nFuturVal>,<nRate>,<nPeriods>)
     Example.....: ? _app.Math.Discount(1000,.08,6)
     Returns.....: Numeric
     Parameters..: nFuturVal = the amount to be received/paid in the 
                               future
                   nRate     = the periodic rate of interest
                   nPeriods  = the number of periods
    ------------------------------------------------------------------- */

   RETURN ((nFuturVal) / ( 1 + (nRate) ) ^ (nPeriods) )
   // -- EoM: Discount()

   FUNCTION Factorial( nNumber )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Factorial of a number; returns -1 if number is not
                   a positive integer.
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Factorial(<nNumber>)
     Example.....: ? _app.Math.Factorial( 6 )
     Returns.....: Numeric = number factorial <in example, 6! or 720>
     Parameters..: nNumber = number for which factorial is to be 
                             determined
    ------------------------------------------------------------------- */
      
      local nNext, nProduct
      
      if nNumber # this.int( nNumber ) or nNumber < 1
         RETURN -1
      endif
      nProduct = 1
      nNext = nNumber
      for nNext = nNumber to 2 step - 1
          nProduct = nProduct * nNext
      endfor
      
   RETURN nProduct
   // -- EoM: Factorial()

   FUNCTION FuturVal( nPresVal, nRate, nPeriods )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Compute the future value of an initial amount at 
                   compound interest received at a given periodic 
                   rate for a number of periods.
     Written for.: dBASE IV, 1.0
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.   
     Calls.......: None
     Usage.......: math.FuturVal(<nPresVal>,<nRate>,<nPeriods>)
     Example.....: ? _app.Math.FuturVal(10000,.06,48)
     Returns.....: Numeric
     Parameters..: nPresVal = Present Value
                   nRate    = Periodic interest rate
                   nPeriods = Number of periods to calculate for
    ------------------------------------------------------------------- */

   RETURN ((nPresVal) * ( 1 + (nRate)) ^ (nPeriods))
   // -- EoM: FuturVal()

   FUNCTION FVIrr( aFlow, lSw, nGuess )
   /* -------------------------------------------------------------------
     Programmer..: Ron Allen 
     Date........: 11/09/1996
     Notes.......: Returns same roots as Irr2(), but averages 20% 
                   faster. Irr2() searches for the roots of NPV (Net 
                   Present Value), while FVirr() searches for the 
                   same roots of NFV (Net Future Value), both with 
                   respect to the rate of return. The user may wish 
                   to use this UDF in place of Irr2() and use Irr2() 
                   as an alternate to help locate more multiple 
                   solutions. The reason this UDF is 'usually' faster
                   is due to the fact that the NFV curve is 'usually'
                   steeper as it crosses the zero axis.
     Written for.: dBASE IV, 1.5
     Rev. History: 01/28/1993 -- Original
                   01/28/1993 -- 1.01 - Modified Irr() to use Net 
                   Future Value curve instead of Net Present Value 
                   curve.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.   
     Calls.......: None
     Usage.......: math.FVIrr( <aFlow>, <lSw>, <nGuess>)
     Example.....: Rate = _app.Math.FVIrr( Cash, Switch, .01)
     Returns.....: Internal Rate of Return.
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should be one-dimensioned and have as 
                            many elements as the cashflows, no more. 
                            Do NOT pass the name in quotation marks.
                   lSw    = name of a logical variable to be switched
                            to indicate valid IRR returned (true).
                   nGuess = optional guess for initialing search.
    ------------------------------------------------------------------- */
      
      local nI, nPosVal, nNegVal, nCurVal, nIRR, nNuDelta, nOlDelta, ;
                nBigchange
      local nSignChng, nDiscount, nRatio, nSumFV, nCurrFV, nSumDeriv, ;
                nOldFV
      local nIters, lSw1
      
      store 0 to nPosVal, nNegVal, nIters
      store true to lSw
      store false to lSw1
       
      // -- Separate negatives from positives
      for nI = 1 to alen( aFlow )
          store aFlow[nI] to nCurVal
          if nCurVal < 0
             nNegVal = nNegVal + nCurVal
          else
             nPosVal = nPosVal + nCurVal
          endif
      endfor
      if nNegVal = 0 .or. nPosVal = 0
         wait "Must have at least one positive and one negative value"
         lSw1 = false
         RETURN 0
      endif
      
      /*  Use initializing guess if provided, otherwise calculate from
        -- weighted average returns. */
      if pcount() = 4
         nIRR = nGuess
      else
         nIRR = ((-nPosVal/nNegVal)-1)/nN
      endif
      
      // -- Housekeeping summary accumulators, etc., before entering loop
      store 1 to nNuDelta, nOlDelta
      store 0 to nSignChng, nBigChange
      
      // -- Loop until estimated rate indicated accuracy
      do while this.abs(nNuDelta) > .000001
         store 0 to nSumFV, nSumDeriv
            
         /* -- Set up cumulative denominator to calculate 
            -- incremental NFV */
         nRatio = 1 + nIRR
         nDiscount = nRatio^nN
         for nI = 1 to alen( aFlow )
             nDiscount = nDiscount/nRatio
                
             // -- Calculate incremental FV and add to sum
             nCurrFV = nDiscount * aFlow[nI]
             nSumFV = nSumFV + nCurrFV
                
             /*  Add incremental first derivative to derivative 
               -- sum */
             nSumDeriv = nSumDeriv - nI * nCurrFV
         endfor
         
         // -- count iterations and test for sign change of future value
         if .not. lSw1 .and. nIters > 0
            lSw1 = iif(sign(nOldFV) = sign(nSumFV),false,true)
         endif
         nIters ++
         nOldFV = nSumFV
         
         // -- Calculate indicated change in IRR
         nNuDelta = nRatio * nSumFV/nSumDeriv
         
         /* -- Test for big changes in adjusted IRR, limit to 10 times
            -- current guess for IRR and count big changes. */
         if abs(nNuDelta/nIRR) > 10
            nNuDelta = sign(nNuDelta) * 10 * nIRR
            nBigChange ++
         endif
         
         // -- Make adjustment to guess for IRR
         nIRR = nIRR - nNuDelta
         
         // -- Count reversals in adjustments to limit hunting
         nSignChng = nSignChng + iif(sign(nNuDelta) +;
                     sign(nOlDelta) = 0,1,0)
         nOlDelta = nNuDelta
            
         /* -- Test for hunting, too many bigchanges or too large a 
            -- solution and set external switch if abnormal exit is 
            -- used. */
         if nSignChng + nBigChange > 10 or this.abs(nIRR) >;
               100 or (nIters > 9 and not lSw1)
            store false to lSw
            exit
         endif
      enddo
       
   RETURN nIRR
   // -- EoM: FVIrr()

   FUNCTION FVXIrr( aFlow, lSw, nGuess )
   /* -------------------------------------------------------------------
     Programmer..: Ron Allen
     Date........: 11/09/1996
     Notes.......: Same as XIrr() except that the Net Future Value 
                   (NFV) function is used instead of the Net Present 
                   Value (NPV) function. The roots are the same, but 
                   this function is usually faster for the same 
                   reasons that FVirr() is faster than Irr2(). As in
                   Xirr(), all dates except the first date in the 
                   array may be in random order. The first date in 
                   the nN by 2 array along with the maximum date 
                   establishes the range of the investment analysis. 
     Written for.: dBASE IV, 1.5
     Rev. History: 01/28/1993
                   01/28/1993 -- 1.01 - Modified XIrr() to find roots
                     of the Net Future Value curve.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.FVXIrr( <aFlow>, <lSw>, <nGuess>)
     Example.....: Rate = _app.Math.FVXIrr( Cash, Switch, .01)
     Returns.....: Effective Internal Rate of Return.
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should two-dimensioned and have as 
                            many rows as the cashflows, no more. It 
                            should contain the dates in column 1 and 
                            the cashflow amounts in column 2. Do NOT
                            pass the name in quotation marks.
                   lSw    = name of a logical variable to be switched 
                            to indicate valid IRR returned (true).
                   nGuess = optional guess for initializing search.
    ------------------------------------------------------------------- */
      
      local nI, nPosVal, nNegVal, nCurVal, nIRR, nNuDelta, nOlDelta, ;
                 nBigchange
      local nSignChng, nRatio, dPDate, dMaxDate, nCurrFV, nSumDeriv
      local nSumFV, lSw1, nIters
      
      store 0 to nPosVal, nNegVal, nIters
      store true to lSw
      store aFlow[1,1] to dMaxDate, dPDate
       
      // -- Separate negatives from positives and find last date.
      
      for nI = 1 to alen( aFlow )
          store aFlow[nI,2] to nCurVal
          store this.max( aFlow[nI,1],dMaxDate) to dMaxDate
          if nCurVal < 0
             nNegVal += nCurVal
          else
             nPosVal += nCurVal
          endif
      endfor
      if nNegVal = 0 or nPosVal = 0
         wait "Must have at least one positive and one negative value"
         lSw = false
         RETURN 0
      endif
      
      /* -- Use initializing guess if provided, otherwise calculate from
         -- weighted average returns. */
      if pcount() = 4
         nIRR = nGuess
      else
         nIRR = (((nPosVal+nNegVal-aFlow[1,2])/-;
                    aFlow[1,2])-1)/(dMaxDate-dPDate)
      endif
       
      // -- Housekeeping summary accumulators, etc., before entering loop
      store 1 to nNuDelta, nOlDelta
      store 0 to nSignChng, nBigChange
      store false to lSw1
      
      // -- Loop until estimated rate indicated accuracy
      do while abs(nNuDelta) > .000001
         store 0 to nSumFV, nSumDeriv
         store 1 + nIrr to nRatio
         for nI = 1 to alen( aFlow )
             
             // -- Calculate incremental FV and add to sum
             nCurrFV =  aFlow[nI,2] * nRatio^;
                       (dMaxDate - aFlow[nI,1])
             nSumFV = nSumFV + nCurrFV
             
             /*  Add incremental first derivative to derivative 
               -- sum */
             nSumDeriv = nSumDeriv + (dMaxDate - ;
                         aFlow[nI,1]) * nCurrFV
         endfor
          
         // -- count iterations and test for sign change of future value
         if not lSw1 and nIters > 0
            lSw1 = iif(sign(nOldFV) = sign(nSumFV),false,true)
         endif
         nIters++
         nOldFV = nSumFV
         
         // -- Calculate indicated change in IRR
         nNuDelta = nRatio * nSumFV/nSumDeriv
         
         /* -- Test for big changes in adjusted IRR, limit to 10 times
            -- current guess for IRR and count big changes. */
         if abs(nNuDelta/nIRR) > 10
            nNuDelta = sign(nNuDelta) * 10 * nIRR
            nBigChange++
         endif
         
         // -- Make adjustment to guess for IRR
         nIRR = nIRR - nNuDelta
           
         // -- Count reversals in adjustments to limit hunting
         nSignChng = nSignChng + iif(sign(nNuDelta) + ;
                          sign(nOlDelta) = 0,1,0)
         nOlDelta = nNuDelta
         
         /*  Test for hunting, too many bigchanges or too large a 
           -- solution and set external switch if abnormal exit is 
           -- used. */
         if nSignChng + nBigChange > 10 or this.abs(nIRR) > 100;
           or (nIters > 9 and not lSw1)
           store false to lSw
           exit
         endif
      enddo
      
   RETURN (1+nIrr)^365 -1
   // -- EoM: FVXIrr()

   FUNCTION GCD( n1, n2 )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Greatest common divisor of two integers.  Given 
                   two integers, returns their largest common 
                   divisor. Returns 0 if one or both are not 
                   integers, but returns the absolute value of the 
                   GCD if one or both are negative.
                   If one is 0, returns the other.
                       Usually known as "Euclid's algorithm."
                       The algorithm used is discussed in 4.5.2 of
                   Volume II, "The Art of Computer Programming", 2d
                   edition, Addison-Wesley, Reading, MA, by Donald 
                   Knuth.
     Written for.: dBASE IV, 1.1 and 1.5
     Rev. History: 12/03/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.GCD( <n1>, <n2> )
     Example.....: ? _app.Math.GCD( 24140, 40902 )
     Returns.....: numeric, the GCD, or 0 if not both integers ( 34 
                   in example).
     Parameters..: n1       = numeric, one of the integers
                   n2       = numeric, the other
    ------------------------------------------------------------------- */
      
      local nMin, nMax, nMod
      
      nMax = iif( int( n1 ) = n1 and int( n2 ) = n2, 1, 0 )
      
      if nMax # 0
         nMin = this.min( this.abs( n1 ), this.abs( n2 ) )
         nMax = this.max( this.abs( n1 ), this.abs( n2 ) )
         
         do while nMin > 0
            nMod = this.mod( nMax, nMin )
            nMax = nMin
            nMin = nMod
         enddo
      endif
       
   RETURN nMax
   // -- EoM: GCD()

   FUNCTION Hav(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Haversine of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Hav(<n>)
     Example.....: ? _app.Math.Hav(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( ( 1 - this.cos(n) ) / 2 )
   // -- EoM: Hav()

   FUNCTION Irr2( aFlow, lSw, nGuess )
   /* -------------------------------------------------------------------
     Programmer..: Ron Allen
     Date........: 11/09/1996
     Notes.......: Returns internal rate of return on an investment 
                   from evenly-spaced periodic cashflows. The UDF 
                   simultaneously accumulates the periodic Net 
                   Present Values of the individual cashflows along 
                   with the first derivative of the function. After 
                   the summation is completed for each guess, the 
                   guess is adjusted by subtracting the ratio of the 
                   function to its derivative. Size the array to the 
                   number of cashflows before calling the function.
     Written for.: dBASE IV, 1.5
     Rev. History: 01/25/1993 -- Original
                   01/28/1993 - 1.01 -- to add missing private 
                   variables. To count iterations without sign change
                   in PV. Move division by nRatio outside inner loop.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Irr2(<aFlow>, <lSw>[, <nGuess>] )
     Example.....: Rate = _app.Math.Irr2( Cash, Switch, .01)
     Returns.....: Internal Rate of Return.
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should be one-dimensioned and have as 
                            many elements as the cashflows, no more. 
                            Do NOT pass the name in quotation marks.
                   lSw    = name of a logical variable to be switched
                            to indicate valid IRR returned (true).
                   nGuess = optional guess for initialing search.
    ------------------------------------------------------------------- */
      
      local nI, nPosVal, nNegVal, nCurVal, nIRR, nNuDelta, nOlDelta, ;
                 nBigchange
      local nSignChng, nDiscount, nRatio, nSumPV, nCurrPV, nSumDeriv, ;
                 nOldPV
      local nIters, lSw1, nSize
      store 0 to nPosVal, nNegVal, nIters
      store true to lSw
      store false to lSw1
      
      // --  separate positives from negatives
      
      for nI = 1 to alen( aFlow )
          store aFlow[nI] to nCurVal
          if nCurVal < 0
             nNegVal += nCurVal
          else
             nPosVal += nCurVal
          endif
      endfor
      if nNegVal = 0 or nPosVal = 0
         wait "Must have at least one positive and one negative value"
         store false to lSw
         RETURN 0
      endif
      
      /* -- Use initializing guess if provided, otherwise calculate from
         -- weighted average returns. */
      
      if pcount() = 4
         nIRR = nGuess
      else
         nIRR = ((-nPosVal/nNegVal)-1)/nN
      endif
      
      // -- Housekeeping summary accumulators, etc., before entering loop
      store 1 to nNuDelta, nOlDelta
      store 0 to nSignChng, nBigChange
      
      // --  Loop until estimated rate indicated accuracy
      do while abs(nNuDelta) > .000001
         store 0 to nI, nSumPV, nSumDeriv
         
         /*  Set up cumulative denominator to calculate 
           -- incremental NPV */
         nDiscount = 1
         nRatio = 1 + nIRR
         do while nI < nN
            nI++
            nDiscount = nDiscount/nRatio
               
            // -- Calculate incremental PV and add to sum
            nCurrPV = nDiscount * aFlow[nI]
            nSumPV += nCurrPV
            
            /* -- Add incremental first derivative to derivative 
               -- sum */
            nSumDeriv = nSumDeriv - nI * nCurrPV
         enddo
         
         // -- count iterations and test for sign change of future value
         if not lSw1 and nIters > 0
            lSw1 = iif(sign(nOldPV) = sign(nSumPV),false,true)
         endif
         nIters++
         nOldPV = nSumPV
         
         // -- Calculate indicated change in IRR
         nNuDelta = nRatio * nSumPV/nSumDeriv
         
         /* -- Test for big changes in adjusted IRR, limit to 10 times
            -- current guess for IRR and count big changes. */
         if abs(nNuDelta/nIRR) > 10
            nNuDelta = sign(nNuDelta) * 10 * nIRR
            nBigChange++
         endif
         nIRR -= nNuDelta   /*  Make adjustment to guess
                                for IRR */
         
         // -- Count reversals in adjustments to limit hunting
         nSignChng = nSignChng + iif(sign(nNuDelta) + ;
                     sign(nOlDelta) = 0,1,0)
         nOlDelta = nNuDelta
         
         /* -- Test for hunting, too many bigchanges or too large a 
            -- solution and set external switch if abnormal exit is 
            -- used. */
         if nSignChng + nBigChange > 10 or this.abs(nIRR) > 100 ;
            or (nIters > 9 and not lSw1)
            store false to lSw
            exit
         endif
      enddo
      
   RETURN nIRR
   // -- EoM: Irr2()

   FUNCTION isPrime( nNumber )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Returns true if argument is prime positive integer,
                   or false
     Written for.: dBASE IV, 1.1
     Rev. History: 03/11/1992 -- original function.
                 : 08/11/1992 -- revised to return true for 2.
                   09/28/1994 -- Jay Parsons, revised to use for loop
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: isPrime(<nNumber>)
     Example.....: ? isPrime( 628321 )
     Returns.....: Logical = true if prime
     Parameters..: nNumber = positive integer to test for being prime
    ------------------------------------------------------------------- */
      
      local nFactor, lResult
      
      if nNumber < 1 or nNumber # this.int( nNumber ) ;
         or ( nNumber > 2 AND this.mod( nNumber, 2 ) = 0 )
         RETURN false
      endif
      lResult = true
      for nFactor = 3 to this.sqrt( nNumber ) step 2
          if this.mod( nNumber, nFactor ) = 0
             lResult = false
             exit
          endif
      endfor
      
   RETURN lResult
   // -- EoM: isPrime()

   FUNCTION MIrr( aFlow, nRrate, nFrate )
   /* -------------------------------------------------------------------
     Programmer..: Ron Allen
     Date........: 11/09/1996
     Notes.......: Used to calculate the Modified Internal Rate of 
                   Return for evenly-spaced periodic cashflows. The 
                   modifications assume that more realistic 
                   investment models should account for the cost of 
                   borrowing or the lower 'safe' rate for keeping 
                   reserve funds to cover outlays and the fact that 
                   reinvestments will be made at some other rate than
                   the IRR itself. This model calculates the answer 
                   directly, therefore more rapidly than the 
                   iterative approach used by IRR. 
     Written for.: dBASE IV, 1.5
     Rev. History: 01/27/1993 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: MIrr( <aFlow>, <nRrate>, <nFrate>)
     Example.....: Rate = MIrr( Cash, .1, .14)
     Returns.....: Modified Internal Rate of Return per period.
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should be one-dimensioned and have as 
                            many elements as the cashflows, no more. 
                            Do NOT pass the name in quotation marks.
                   nRrate = Reinvestment rate for positive cashflows. 
                   nFrate = 'Safe' rate expected on reserve funds to 
                             cover disbursements.
    ------------------------------------------------------------------- */
    
      local nI, nNegVal, nPosVal, nCurVal
      store 0 to nNegVal, nPosVal
      
      /* -- Pass through array once computing present value of negative
         -- cashflows at 'safe' rate and present value of positive values
         -- at the reinvestment rate. */
      for nI = 1 to alen( aFlow )
          nCurVal = aFlow[nI]
          nCurVal = nCurVal*(1+iif(nCurVal<0,nFrate,nRrate))^;
                    -(nI-1)
          if nCurVal < 0
             nNegVal += nCurVal
          else
             nPosVal += nCurVal
          endif
      endfor
      if this.abs(nNegVal) = 0 or nPosVal = 0
         wait ;
          "There must be at least one negative and one positive value!"
         RETURN 0
      endif
      
      /* -- Calculate the rate of return required to yield a future value
         -- of the positive values reinvested at nRrate from the present
         -- value of the negative values invested at the 'safe' rate. */
       
   RETURN ((-nPosVal * (1+nRrate)^(nN-1))/nNegVal)^;
          (1/(nN-1))-1
   // -- EoM: MIrr()

   FUNCTION MultDice( nNum, nSides )
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 11/10/1996
     Notes.......: Function used to determine a random total for the
                   roll of multiple dice, of x# of sides.
     Written for.: dBASE IV, 1.1
     Rev. History: 06/12/1991 -- original function.
                   02/13/1992 -- cleaned up to call DICE for each
                    iteration, rather than calling once and then 
                    redoing the randomizer logic ... 
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/10/1996 -- Modified to work with this class.
     Calls.......: mathEx.Dice()
     Usage.......: mathex.MultDice(<nNum>,<nSides>)
     Example.....: nVal = _app.Math.MultDice(3,6)
     Returns.....: Random value of 1 to x (x being number of sides), 
                   for each iteration (nNum), totalled. For example,
                   value returned would be the total of 3 six-sided 
                   die rolled, the number would be anywhere from 3 to
                   18.
     Parameters..: nNum   = Number of dice to be "rolled"
                   nSides = # of sides to the dice (see Dice() above)
    ------------------------------------------------------------------- */
      
      local nCount, nTotal
      
      nTotal = 0                             // set total
      for nCount = 1 to nNum                 // loop for number of dice
          nTotal = nTotal + CLASS::Dice(nSides) // add to total
      endfor
         
   RETURN nTotal
   // -- EoM: MultDice()

   FUNCTION Npv2( aName, nRate )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Net present value of array aName given assumed 
                   nRate. This is a rewrite of Npv, by Tony Lima and 
                   Jay Parsons, renamed because the parameters have 
                   changed.
     Written for.: dBASE 5.0 for Windows
     Rev. History: 10/04/1994 -- Original Release
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Npv2(<aName>,<nRate>)
     Example.....: ? _app.Math.Npv2( aMyArray, .06 )
     Returns.....: Float = value of the project at given rate
     Parameters..: aName, name of the array holding the cashflows.
                   nRate    = Interest Rate
     Other Inputs: The array must be set up with one dimension, each
                   element holding a cashflow. The first cashflow 
                   must be the present investment, all others must be
                   at equal intervals with 0 elements for periods of 
                   no cash. Payments by you are negative, to you 
                   positive. For a project expected to require an 
                   immediate outlay of $6,000 and to return $2,000 at
                   the end of each of the first five years 
                   thereafter, the array will be:
                         aCash[1] = -6000
                         aCash[2] =  2000
                         aCash[3] =  2000
                             * * *
                         aCash[6] =  2000
                   The nRate rate must be stated in terms of the same
                   period as the interval between array elements.
    ------------------------------------------------------------------- */
      
      local nDiscount, nFactor, nPeriod, nNpv
      
      nNPV = aName[ 1 ]
      nDiscount = float( 1 )
      nFactor = 1 / ( 1 + nRate )
      for nPeriod = 2 to alen( aName )
          nDiscount = nDiscount * nFactor
          nNPV = nNPV + aName[ nPeriod ] * nDiscount
      endfor
      
   RETURN nNPV
   // -- EoM: Npv2()

   FUNCTION ParseDice( cDice )
   /* -------------------------------------------------------------------
     Programmer...: Ken Mayer 
     Date.........: 11/10/1996
     Notes........: This is another gaming function ...
                    Its purpose is to read a string in the format  
                    xdy+z or some variation, and calculate the value
                    x = # of dice, 
                    d = a part of the standard gaming syntax (i.e., 
                        3d6),
                    y = # of sides of dice,
                    + = a modifier (could be a minus also ...)
                    z = number to modify each die rolled
                    (3d6+1 = a value from 6 to 21 (figure if you add 
                    1 to each die rolled, minimum value will be 6 
                    (3+3), maximum will be 21 (18+3))).)
     Written for.: dBASE IV, 1.1
     Rev. History: 08/29/1991 -- original function.
                   02/13/1992 -- minor -- changed randomizer call to 
                     DICE()
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/10/1996 -- Modified to work with this class.
     Calls.......: string.AllTrim()
                   mathex.Dice()
     Usage.......: mathex.ParseDice("<cDice>")
     Example.....: ? _app.Math.ParseDice("5d6-3")
     Returns.....: Random number from x (modified by z) to y 
                   (modified by z)
     Parameters..: cDice = Standard gaming format value to be parsed 
                           and calculated.
    ------------------------------------------------------------------- */
      
      local nCount,cDice,nPos,nNumDice,nMod,nDice,nPos2,nReturn
      
      cDice = upper( this.String.AllTrim(cDice) )
         
      if at("D",cDice) > 0      // if the letter 'D' is in there
         /* -- get the VALUE of the "substring" of m->cDice, starting at
            -- character 1, going to the letter D and backing up 1.
            -- this will be useful in case we have 10dy ... otherwise,
            -- we _could_ assume only one character, but assumptions are
            -- bad ... */
         nPos = at("D",cDice)
         nNumDice = val(substr(cDice,1,nPos-1))
         nPos++       // move to character beyond letter 'D'
         if at("+",cDice) > 0   // if we have a + modifier
            nPos2 = at("+",cDice)
            nDice = val(substr(cDice,nPos,nPos2-1))
            nMod = val(substr(cDice,nPos2+1,len(cDice)-nPos2))
         else
            if at("-",cDice) > 0 // if we have a - modifier
               nPos2 = at("-",cDice)
               nDice = val(substr(cDice,nPos,nPos2-1))
               nMod = val(substr(cDice,nPos2+1,;
                                      len(cDice)-nPos2))
            else  // no modifier
               nDice = val(substr(cDice,nPos,;
                                      len(cDice)-nPos+1))
            endif  // check for - sign
         endif  // check for + sign
         
         // -- roll the nDice sided "dice" nNumDice number of times
         nReturn = 0
         for nCount = 1 to nNumDice
             nReturn = nReturn + CLASS::Dice(nDice)
         endfor
            
         // -- Modifiers -- add or subtract appropriate value
         if at("+",cDice) > 0  // if there's a + sign,
            nReturn = nReturn + (nNumDice * nMod)
         endif
         if at("-",cDice) > 0  // it's a minus sign
            nReturn = nReturn - (nNumDice * nMod)
         endif
            
      else  /* there's no letter 'D', so we simply have a number to
               return - this is under the assumption that the value
               passed is either a random one, or (in this case)
               it's a set value ... for example, in some cases
               in my gaming system, HitPoints for a critter may
               be a set value, in others it may be a random one.
               this routine handles both ... */
            
         nReturn = val(cDice)
            
      endif
      
   RETURN nReturn
   // -- EoM: ParseDice()

   FUNCTION Permutes( nNum, nHowMany )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Permutations of nNum items taken Nhowmany at a 
                   time That is, the number of possible arrangements, 
                   as the different ways a president, V.P. and sec'y 
                   may be chosen from a club of 10 members
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   11/09/1996 -- Modified to work with this class.   
     Calls.......: None
     Usage.......: Permutes(<nNum>,<nHowMany>)
     Example.....: ?Permutes(10,3)
     Returns.....: Numeric
     Parameters..: nNum     = number of items in the entire set
                   nHowMany = number to be used at once
    ------------------------------------------------------------------- */
      
      local nResult, nCounter
      store 1 to nResult
      for nCounter = 1 to nHowmany
         nResult = nResult * ( nNum + 1 - nCounter )
      endfor
      
   RETURN nResult
   // -- EoM: Permutes()

   FUNCTION RandSel( nN, nT, aName, lReseed )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/10/1996
     Notes.......: Random selection of integers.  The function 
                   requires two numeric parameters, the number nN to 
                   select and the number of items nT to select from.  
                   It fills the first nN rows of a one-column array 
                   with an ordered random selection of the integers 
                   from 1 to nT, which may of course be used as 
                    record numbers or indices into some other data 
                   structure to select items from it.  If passed a 
                   third, array, parameter, it will place the 
                   selected numbers in the array of that name, 
                   otherwise in the array "RandSel".  If passed a 
                   fourth parameter that evaluates to true, it will 
                   reseed the random number generator, otherwise use 
                   the next random numbers.
                       If the array does not exist, it will be 
                   created. If it does exist but with two dimensions 
                   or too few rows, it will be resized. If the first 
                   parameter is larger than the second, they will be 
                   swapped.
                       The random-number generator should usually be 
                   reseeded, either by using the "lReseed" parameter 
                   or before calling the function, except where the 
                   function is being called repeatedly either within 
                   a very short time or for related applications in 
                   which a repetition of the sequence would defeat 
                   the randomness.
                       For dBASE IV versions before 1.5, revise this 
                   to take only the two numeric parameters by 
                   commenting out the first "parameters" line of code 
                   below and including the next three commented 
                   lines. The array "RandSel" will be used, and 
                   reseeding if needed must be done before calling 
                   the function.
                       The algorithm used is "Algorithm S" discussed
                   in 3.4.2 of Volume II, "The Art of Computer 
                   Programming", 2d edition, Addison-Wesley, Reading,
                   MA, by Donald Knuth.
     Written for.: dBASE IV, 1.1 and 1.5
     Rev. History: 12/03/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/10/1996 -- Modified to work with this class.   
     Calls.......: None
     Usage.......: mathex.RandSel( <nN>,<nT> [,<aName> [,<lReseed>]])
     Example.....: lX = _app.Math.RandSel( 100, reccount(),;
                                             MyArray,true)
     Returns.....: true if successful, or false if given number < 1 as 
                   parameter.
     Parameters..: nN       = numeric, number of integers to select
                   nT       = numeric, highest integer to select from
                   aName    = name, not in quotes, of the array to 
                              hold the selected integers.  If not 
                              furnished, array "RandSel" will be 
                              used.
                   lReseed  = logical, true to reseed the random-
                              number generator.  Default is false, no 
                              reseed.
     Side effects: Creates as needed and fills the array.
                   Uses some random numbers from the sequence.
    ------------------------------------------------------------------- */
      
      local nChoose, nTotal, lReturn, nChosen, nSeen
      
      nChoose = int( min( nN, nT ) )
      nTotal = int( max( nN, nT ) )
      lReturn = ( nChoose >= 1 )
      
      if lReturn
         if type( "aName" ) # "A"
            aName = "RandSel"
            RandSel = new Array(  nChoose )
         else
            if alen( aName, 2 ) > 0 or alen( aName, 1 ) # nChoose
               lReturn = ( aresize( aName, nChoose, 0 ) = nChoose )
            endif
         endif
      endif
      
      if lReturn
         if lReseed
            nX = this.random( -1 )
         endif
         store 0 to nChosen, nSeen
         do
            if this.random() * ( nTotal - nSeen ) < nChoose - nChosen
               nChosen++
               aName.add( nSeen + 1 )
            endif
            nSeen++
         until nChosen = nChoose
      endif

   RETURN lReturn
   // -- EoM: RandSel()

   FUNCTION Rate( nFutVal, nPresVal, nPeriods )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Compute rate of periodic interest needed to 
                   produce a future value from a present value in a 
                   given number of periods.  If the periods are not 
                   years, you'll probably want to multiply the rate 
                   returned by the number of periods in a year to 
                   obtain the equivalent annual rate.
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Rate(<nFutVal>,<nPresVal>,<nPeriods>)
     Example.....: ? _app.Math.Rate(50000,10000,48)
     Returns.....: Numeric
     Parameters..: nFutVal  = Future Value
                   nPresVal = Present Value
                   nPeriods = Number of periods to calculate for
    ------------------------------------------------------------------- */

   RETURN ( ( (nFutVal)/(nPresVal) ) ^ ( 1 / (nPeriods) ) - 1 )
   // -- EoM: Rate()

   FUNCTION RMS( cField )
   // NOTE: This method assumes you're using the xBase DML, 
   // not the database classes:
   /* -------------------------------------------------------------------
     Programmer..: Oktay Amiry 
     Date........: 11/09/1996
     Notes.......: Root-Mean-Square can be applied to any numeric 
                   list (ordinal, interval, and ratio) to find the 
                   overall size of the numbers in the list, ignoring 
                   their signs.
                   Printed in TechNotes, December 1992.
     Written for.: dBASE IV, 1.5
     Rev. History: 12/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.RMS(<cField>)
     Example.....: ? oMath.RMS("SCORE")
     Returns.....: numeric
     Parameters..: cField = a numeric field in the current table
    ------------------------------------------------------------------- */
      
      local nTotal
      
      calculate sum(&nField. ^ 2) to nTotal
      
   RETURN sqrt((nTotal/reccount()))
   // -- EoM: RMS()

   FUNCTION SampleVar( cField, cClause )
   // NOTE: This method assumes you're using the xBase DML, 
   // not the database classes:
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Finds sample variance of specified field of the 
                   current database, using CALCULATE command.
                   The CALCULATE command calculates the population 
                   variance, which is smaller by a factor of (n-1)/n.
     Written for.: dBASE IV Version 1.5
     Rev. History: Original function 1990.
                   Modified to take optional parameter, 4/13/1992
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
     Calls.......: None
     Usage.......: SampleVar( <cField> [, <cClause> ] )
     Example.....: ? SampleVar( "Balance", ".FOR..NOT.;
                                           isblank( Balance )" )
     Returns.....: a numeric or float value, the sample variance, or
                   false if it cannot be calculated. If any of the 
                   numeric items are floats, the result will be.
     Parameters..: cField  = name of a numeric field of the current 
                             database for which to calculate the 
                             sample variance
                   cClause = optional, a FOR, WHILE, TO, etc. clause
    ------------------------------------------------------------------- */
      
      local fVar, nCount, cCond
      
      if pcount() = 2
         cCond = " "+ cCondition
      else
         cCond = ""
      endif
      calculate var( &cField. ), CNT() TO fVar, nCount &cCond.
       
   RETURN iif( nCount > 1, fVar * nCount / ( nCount - 1 ), false )
   // -- EoM: SampleVar()

   FUNCTION SD( cField, cType )
   // NOTE: This method assumes you're using the xBase DML, 
   // not the database classes:
   /* -------------------------------------------------------------------
     Programmer..: Oktay Amiry
     Date........: 11/09/1996
     Notes.......: Standard Deviation -- similar to the dBASE STD 
                   function. The standard deviation shows how far 
                   away numbers on a list are from their average. The 
                   value yielded by standard deviation is in the same
                   units as the numbers which are used to calculate 
                   the SD. The SD() function can take two forms: an
                   unbiased (n-1) method and the biased (n-method) 
                   form. The SD() function, by default, takes the 
                   biased form, which is the standard deviation for
                   a population based on the entire population. With 
                   the explicit second parameter being "S", the SD()
                   performs the unbiased method, which is the 
                   standard deviation for a population that is based 
                   on a sample. This latter method, which is 
                   also referred to as the SD+, is usually the value 
                   produced by statistical calculators and is 
                   frequently higher than population-based SD.
                   Printed in TechNotes, December 1992.
     Written for.: dBASE IV, 1.5
     Rev. History: 12/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.SD(<nField>[,"S"])
     Example.....: ? _app.Math.SD("SCORE","S")
     Returns.....: numeric
     Parameters..: cField = a numeric field in the current table
    ------------------------------------------------------------------- */
      
      local nAverage, nEntry
      
      calculate avg(&nField. ^ 2) to nEntry, ;
                avg(&nField.) to nAverage
      nAverage = nAverage ^ 2
      
   RETURN iif(type("CTYPE") = "C" .and. upper(cType) = "S",;
              sqrt(nEntry-nAverage)/sqrt((reccount()-1)/reccount()),;
              sqrt(nEntry-nAverage))
   // -- EoM: SD()

   FUNCTION Secant( n )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Secant of an angle n in radians.
                   Indeterminate and returns error if n is 90 or 270
                       deg., i.e. mod( abs(n) + pi()/2 , pi() ) = 0
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.Secant(<n>)
     Example.....: ? _app.Math.Secant(48)
     Returns.....: Numeric
     Parameters..: n = angle, in radians
    ------------------------------------------------------------------- */

   RETURN ( 1 / this.cos(n) )
   // -- EoM: Secant()

   FUNCTION SecH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Hyperbolic secant of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.CosH()
     Usage.......: math.SecH(<n>)
     Example.....: ? _app.Math.SecH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( 1 / CLASS::CosH(n) )
   // -- EoM: SecH()

   FUNCTION SinH(n)
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Hyperbolic sine of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: SinH(<n>)
     Example.....: ?SinH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ------------------------------------------------------------------- */

   RETURN ( (this.exp(n) - this.exp(-(n))) / 2 )
   // -- EoM: SinH()

   FUNCTION StnArea( nDevs )
   /* -------------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Area of the standard normal distribution function 
                   between mean and given number of standard 
                   deviations from the mean.
                 
                   Knowing that a 650 score is 1.5 standard 
                   deviations from the 500 mean, we can calculate 
                   Stnarea( 1.5 ) as .4332. This tells us that 43.32% 
                   of the scores lie between 650 and 500.  Since 50% 
                   lie below 500, a score of 650 beats 93.32% of the 
                   scores.
                 
                   The polynomial approximation used by this function 
                   is said to be accurate to .00001, 1/1000 of one 
                   percent. Remember to SET DECIMALS appropriately to
                   view results.
                 
     Written for.: dBASE IV
     Rev. History: 11/13/1990 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls       : math.StnY()
     Usage.......: math.StnArea( <nDevs> )
     Example.....: ? _app.Math.StnArea( 1.5 )
     Returns     : % of area between deviations given and the mean, 
                     0<=a<.5.
     Parameters..: nDevs = standard deviations from the mean
    ------------------------------------------------------------------- */
      
      local nX, nV
      
      nX = this.abs( nDevs )
      nV =  1 / ( 1 + .33267 * nX )
      
   RETURN .5 - CLASS::StnY( nX ) * ( .4361836  * nV - .1201676 * nV *;
           nV + .937298 * nV * nV * nV )
   // -- EoM: StnArea()

   FUNCTION StnDevs( nArea )
   /* -----------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Calculates "z", standard deviations, corresponding
                   to any area of standard normal curve toward mean 
                   from the desired z. Much slower than Stnz().
     Written for.: dBASE IV Version 1.5
     Rev. History: Original function 1990.
                   Conformed to Zeroin() 4/13/1992.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302 
                   11/09/1996 -- Modified to work with this class.
     Calls.......: ROOT.CC
     Usage.......: mathex.StnDevs( <nArea> )
     Example.....: ? _app.Math.StnDevs( .96 )
     Returns.....: z, number of standard deviations from mean, or a 
                   negative number indicating failure to find a 
                   root...
     Parameters..: nArea = area "left" of point of interest, 
                            .5 < nArea < 1
    ----------------------------------------------------------------- */
      
      local nTest
      
      if nArea > .99999 OR nArea < .5
         RETURN -1
      endif

      // -- use new Root class:
      this.Root.fp         = this.StnArea
      this.Root.x1         = 0
      this.Root.x2         = 4.2
      this.Root.offset     = nArea - .5
      this.Root.maxiters   = 100
      this.Root.iterations = 0
      this.Root.flag       = 0
      this.Root.abserror   = .000001
      nTest                = this.Root.FindRoot()
      
   RETURN iif( this.Root.Flag < 3, nTest, -this.Root.Flag )
   // -- EoM: StnDevs()

   FUNCTION StnDiff( nConf, nTails, nSample, nSampMean, ;
                     nPopMean, nPopStd )
   /* -----------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Determines whether hypothesis that sample of a 
                   given mean is different from expected mean is 
                   justified.
    
                   if nPopstd, the standard deviation of the 
                   population, is not known and nSample, the sample 
                   size, is greater than 30, the sample standard 
                   deviation may be used for nPopstd.
    
                   This function assumes the population is large 
                   relative to the sample or that the sampling is 
                   with replacement. if neither is true, the right 
                   side of the expression in the later return line 
                   should be multiplied by:
                         sqrt( ( nPop - nSample ) / ( nPop - 1 ) )
                   where nPop is the size of the population.
    
                   Do not use this with small samples, less than 20,
                   because the standard normal distribution is not 
                   sufficiently accurate as an approximation of the 
                   distribution of sample means in such a case.  
                   See "Student's T-distribution" in a statistics 
                   text.
    
     Written for.: dBASE IV Version 1.5
     Rev. History: 04/13/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.StnZ()
     Usage.......: math.StnDiff( <nConf>, <nTails>, <nSample>, ;
                                 <nSampMean>, <nPopMean>, <nPopStd> )
     Example.....: ? _app.Math.StnDiff( .95, 1, 30, 1150, 1200, 20 )
     Returns.....: true if hypothesis of difference is justified to 
                   degree of confidence specified, or false  Returns -1
                   if confidence is not one for which z can be looked
                   up in StnZ().  If you need other confidence 
                   levels, run StnDevs() to find the z values for 
                   them and add them to the StnZ() table.
     Parameters..: nConf     = confidence desired, 0 < nConf < 1
                   nTails    = 1 or 2 = number of tails of curve of 
                               interest
                   nSample   = number of items in the sample
                   nSampMean = mean of the sample
                   nPopMean  = expected mean of the population, the 
                               value which nSampMean should equal, 
                               more or less, if the sample is fair 
                               and the population is as expected  
                   nPopStd   = standard deviation of the population
    ----------------------------------------------------------------- */
      
      local nStd
      
      nStd = CLASS::StnZ( nConf, nTails )
      if nStd = -1
         RETURN nStd
      else
         RETURN this.abs( nSampmean - nPopMean ) ;
                    > nStd * nPopStd / this.sqrt( nSample )
      endif
    // -- EoM: StnDiff()

   FUNCTION StnY( nDevs )
   /* -----------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: Returns value of the standard normal distribution 
                   function given a number of standard deviations 
                   from the mean. This function is not useful alone.  
                   The standard normal distribution function is the 
                   familiar bell-shaped curve scaled so its mean is 
                   at 0, each standard deviation is 1 and the total 
                   area under the curve is 1.  The function StnArea 
                   calls on this function to calculate the 
                   approximate area (a fraction equal to percent of 
                   the total) under the part of the curve lying 
                   between the mean and the given number of standard
                   deviations.
     Written for.: dBASE IV
     Rev. History: 11/13/1990 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls       : None
     Usage.......: math.StnY( <nDevs> )
     Example.....: ? _app.Math.StnY( 1 )
     Returns     : numeric value of the function.
     Parameters..: nDevs = standard deviations from the mean
    ----------------------------------------------------------------- */
 
   RETURN this.exp( -nDevs * nDevs / 2 ) / this.sqrt( 2 * this.pi )
   // -- EoM: StnY()

   FUNCTION StnZ( nProb, nTails )
   /* -----------------------------------------------------------------
     Programmer..: Jay Parsons
     Date........: 11/09/1996
     Notes.......: A lookup table to find the values of "z", standard
                   deviations, corresponding to the most commonly 
                   used areas inside a given number of tails of the 
                   normal distribution function.
                 
                   Used in testing confidence intervals.  If a sample 
                   of light bulbs from a shipment shows an average 
                   life of 1150 hours, and the criterion for 
                   rejection of the shipment is 95% confidence that 
                   the average life of all bulbs is less than (a 
                   single tail) 1200 hours, the value 1.64485 
                   returned by this function is necessary to 
                   determine whether to reject the shipment or not.
                   
                   Values of "z" that are not found in the table can 
                   be found using StnDevs, but it is slow.
     Written for.: dBASE IV
     Rev. History: 11/13/1990 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.StnZ( <nProb>, <nTails> )
     Example.....: ? _app.Math.StnZ( .95, 1 )
     Returns.....: z = number of standard deviations from mean inside
                       which ( or to the side of which includes the 
                       mean, if one tail) the given percentage of 
                       area will fall.
                   Returns -1 if no entry in table.
     Parameters..: nConf  = confidence desired, 0 < nConf < 1
                   nTails = 1 or 2 = number of tails of curve of 
                            interest
    ----------------------------------------------------------------- */
      
      if nTails # 1 and nTails # 2
         RETURN -1
      endif
      do case
         case nConf = .95
              RETURN iif( nTails = 1, 1.64485, 1.96010 )
         case nConf = .99
              RETURN iif( nTails = 1, 2.32676, 2.57648 )
         case nConf = .995
              RETURN iif( nTails = 1, 2.57648, 2.80794 )
         case nConf = .999
              RETURN iif( nTails = 1, 3.09147, 3.29202 )
         otherwise
              RETURN -1
      endcase
    // -- EoM: StnZ()

   FUNCTION TanH(n)
   /* -----------------------------------------------------------------
     Programmer..: Jay Parsons 
     Date........: 11/09/1996
     Notes.......: Hyperbolic tangent of an angle in radians
     Written for.: dBASE IV, 1.1
     Rev. History: 03/01/1992 -- Original Release
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: math.SinH
                   math.CosH	
     Usage.......: math.TanH(<nX>)
     Example.....: ? _app.Math.TanH(48)
     Returns.....: Numeric
     Parameters..: n = angle in radians
    ----------------------------------------------------------------- */
      
   RETURN ( CLASS::SinH(n) / CLASS::CosH(n) )
   // -- EoM: TanH()

   FUNCTION XIrr( aFlow, lSw, nGuess )
   /* -----------------------------------------------------------------
     Programmer..: Ron Allen 
     Date........: 11/09/1996
     Notes.......: Used to calculate the Internal Rate of Return from
                   cashflows on random dates. Except for the need to 
                   supply both the dates of transactions and the 
                   cashflows in an 'nN' by 2 array, the other inputs
                   are the same as in Irr(). Dates may be in random 
                   order except for the first date. The first date in 
                   the array establishes the date to which present 
                   value applies.
     Written for.: dBASE IV, 1.5
     Rev. History: 01/25/1993 -- Original
                   01/28/1993 - 1.01 -- to return 'effective' rate of
                   interest when compounded daily rather than the 
                   'nominal' rate. Also to count iterations without a
                   sign change in PV. Move division by nRatio outside
                   inner loop.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.XIrr( <aFlow>, <lSw>[, <nGuess>] )
     Example.....: Rate = _app.Math.XIrr( Cash, "Switch", .01)
     Returns.....: Effective Internal Rate of Return.
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should two-dimensioned and have as 
                            many rows as the cashflows, no more. It 
                            should contain the dates in column 1 and 
                            the cashflow amounts in column 2. Do NOT
                            pass the name in quotation marks.
                   lSw    = name of a logical variable to be switched 
                            to indicate valid IRR returned (true).
                   nGuess = optional guess for initializing search.
    ----------------------------------------------------------------- */
      
      local nI, nPosVal, nNegVal, nCurVal, nIRR, nNuDelta, nOlDelta, ;
              nBigchange
      local nSignChng, nRatio, dPDate, dMaxDate, nCurrPV, nSumDeriv
      local nSumPV, nIters, lSw1
      
      store 0 to nPosVal, nNegVal, nIters
      store true to lSw
      store aFlow[1,1] to dMaxDate, dPDate
      store false to lSw1
      
      // -- Separate negatives from positives. Find last date.
      for nI = 1 to alen( aFlow, 1 )
          store max( aFlow[nI,1], dMaxDate ) to dMaxDate
          store aFlow[nI,2] to nCurVal
          if nCurVal < 0
             nNegVal += nCurVal
          else
             nPosVal += nCurVal
          endif
      endfor
      if nNegVal = 0 or nPosVal = 0
         wait "Must have at least one positive and one negative value"
         lSw = false
         RETURN 0
      endif
      
      /* Use initializing guess if provided, otherwise calculate from
         weighted average returns. */
      if pcount() = 4
         nIRR = nGuess
      else
         nIRR = (((nPosVal+nNegVal-aFlow[1,2])/-;
                   aFlow[1,2])-1)/(dMaxDate-dPDate)
      endif
      
      // Housekeeping summary accumulators, etc., before entering loop
      store 1 to nNuDelta, nOlDelta
      store 0 to nSignChng, nBigChange
      
      //  Loop until estimated rate indicated accuracy
      do while abs(nNuDelta) > .000001
         store 0 to nSumPV, nSumDeriv
         store 1 + nIrr to nRatio
         for nI = 1 to alen( aFlow, 1 )
                 
             // Calculate incremental PV and add to sum
             nCurrPV =  aFlow[nI,2] / nRatio^;
                        (aFlow[nI,1] - dPDate)
             nSumPV += nCurrPV
             
             /* Add incremental first derivative to derivative 
                sum */
             nSumDeriv = nSumDeriv - (aFlow[nI,1] -;
                             dPDate) * nCurrPV
          endfor
          
          //  count iterations and test for sign change of future value
          if not lSw1 and nIters > 0
             lSw1 = iif(sign(nOldPV) = sign(nSumPV),false,true)
          endif
          nIters++
          nOldPV = nSumPV
             
          //  Calculate indicated change in IRR
          nNuDelta = nRatio * nSumPV/nSumDeriv
             
          /* Test for big changes in adjusted IRR, limit to 10 times
             current guess for IRR and count big changes. */
          if this.abs(nNuDelta/nIRR) > 10
             nNuDelta = sign(nNuDelta) * 10 * nIRR
             nBigChange++
          endif
          //  Make adjustment to guess for IRR
          nIRR -= nNuDelta
             
          // Count reversals in adjustments to limit hunting
          nSignChng = nSignChng + iif(sign(nNuDelta) + ;
                      sign(nOlDelta) = 0,1,0)
          nOlDelta = nNuDelta
          
          /* Test for hunting, too many bigchanges or too large a 
             solution and set external switch if abnormal exit is 
             used. */
          if nSignChng + nBigChange > 10 or this.abs(nIRR) > 100;
             or (nIters > 9 and not lSw1)
             store false to lSw
             exit
          endif
      enddo
      
   RETURN (1+nIrr)^365 -1
   // -- EoM: XIrr()

   FUNCTION XMirr( aFlow, nRrate, nFrate )
   /* -----------------------------------------------------------------
     Programmer..: Ron Allen
     Date........: 11/09/1996
     Notes.......: Used to calculate the Modified Internal Rate of 
                   Return from cashflows on random dates. Except for 
                   the need to supply both the dates of transactions 
                   and the cashflows in an 'nN' by 2 array, the other 
                   inputs are the same as in Mirr(). Dates may be in 
                   random order except for the first date. The first 
                   date in the array establishes the date to which 
                   present value applies. Enter 'Safe' rate for 
                   reserves and 'Reinvestment' rate for positive 
                   cashflows as annual rates, e.g., .075 for 7.5%. 
     Written for.: dBASE IV, 1.5
     Rev. History: 01/27/1993 -- 1.01 - to allow entry of 'Safe' 
                   reserve rate and 'Reinvestment' rate as annual 
                   rates rather than rates. Also, to return the 
                   'effective' rate of interest when compounded 
                   daily, rather than the 'nominal' rate.
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302
                   11/09/1996 -- Modified to work with this class.
     Calls.......: None
     Usage.......: math.XMirr( <aFlow>, <nRrate>, <nFrate>)
     Example.....: Rate = _app.Math.XMirr( Cash, .14, .1)
     Returns.....: Annualized Effective Modified Internal Rate of 
                   Return based on daily compounded interest.   
     Parameters..: aFlow  = name of the array holding the cashflows. 
                            It should two-dimensioned and have as 
                            many rows as the cashflows, no more. It 
                            should contain the dates in column 1 and 
                            the cashflow amounts in column 2. Do NOT
                            pass the name in quotation marks.
                   nRrate = Reinvestment rate for positive cashflows.
                   nFrate = 'Safe' rate expected on reserve funds to 
                             cover disbursements.
    ----------------------------------------------------------------- */
      
      local nI, nCurVal, nNegVal, nPosVal, dPDate, dMaxDate, nCurN, nMirr
      
      store 0 to nNegVal, nPosVal
      store (1+nRrate)^(1/365)-1 to nRrate
      store (1+nFrate)^(1/365)-1 to nFrate
      store aFlow[1,1] to dPDate, dMaxDate
      
      for nI = 1 to alen( aFlow, 1 )
          nCurVal = aFlow[nI,2]
          dCurDate = aFlow[nI,1]
          dMaxDate = max(aFlow[nI, 1],dMaxDate)
          nCurN = dCurDate-dPDate
          nCurVal = nCurVal/(1+iif(nCurVal<0,nFrate,;
                    nRrate))^nCurN
          if nCurVal < 0
             nNegVal += nCurVal
          else
             nPosVal += nCurVal
          endif
      endfor
      if nNegVal = 0 or nPosVal = 0
         wait;
         " There must be at least one negative and one positive value!"
         RETURN 0
      endif
      nN = dMaxDate - dPDate
      nMirr = ((-nPosVal * (1+nRrate)^(nN-1))/nNegVal)^;
              (1/(nN-1))-1
      
   RETURN (1+nMirr)^365-1
   // -- EoM: XMirr()


   FUNCTION ZeroIn( cFunction, fX1, fX2, fAbsError, nMaxIter, n_Flag,;
                    xPass1, xPass2, xPass3 )
   /* -----------------------------------------------------------------
     Programmer..: Tony Lima and Jay Parsons
     Date........: 11/09/1996
     Notes.......: Finds a zero of a continuous function.
                   In substance, what this function does is close in 
                   on a solution to a function that cannot otherwise 
                   be solved. Assuming Y = f(X), if Y1 and Y2, the 
                   values of the function for X1 and X2, have 
                   different signs, there must be at least one value
                   of X between X1 and X2 for which Y = 0, if the 
                   function is continuous. This function closes in on
                   such a value of X by a trial-and-error process.
                 
                   This function is very slow, so a maximum number of
                   iterations is passed as a parameter. If the number
                   of iterations is exceeded, the function will fail 
                   to find a root.  if this occurs, pick different 
                   original "X" values, increase the number of 
                   iterations or increase the errors allowed. Once 
                   an approximate root is found, you can use values 
                   of X close on either side and reduce the error 
                   allowed to find an improved solution. Also, of 
                   course, the signs of Y must be different for the 
                   starting X values for the function to proceed at 
                   all.
                 
                   NOTE ESPECIALLY - There is NO guarantee that a 
                   root returned by this function is the only one, or
                   the most meaningful. It depends on the function 
                   that this function calls, but if that function has
                   several roots, any of them may be returned. This 
                   can easily happen with such called functions as 
                   net present value where the cash flows alternate 
                   from positive to negative and back, and in many 
                   other "real life" cases. See the discussion of 
                   @IRR in the documentation of a good spreadsheet 
                   program such as Quattro Pro for further 
                   information.
                 
                   The method used by this function is a "secant and
                   bisect" search.  The "secant" is the line 
                   connecting two X,Y points on a graph using 
                   standard Cartesian coordinates. Where the secant 
                   crosses the X axis is the best guess for the value 
                   of X that will have Y = 0, and will be correct if 
                   the function is linear between the two points. The 
                   basic strategy is to calculate Y at that value of 
                   X, then keep the new X and that one of the old X 
                   values that had a Y-value of opposite sign, and 
                   reiterate to close in.
                 
                   If the function is a simple curve with most of the
                   change in Y close to one of the X-values, as often
                   occurs if the initial values of X are poorly 
                   chosen, repeated secants will do little to find a 
                   Y-value close to zero and will reduce the 
                   difference in X-values only slightly. In this case 
                   the function shifts to choosing the new X halfway 
                   between the old ones, bisecting the difference and 
                   always reducing the bracket by half, for a while.
                 
                   While this function finds a "zero", it may be used 
                   to find an X corresponding to any other value of 
                   Y. Suppose the function of X is FUNCTION 
                   Blackbox( X ) and it is desired to find a value of 
                   X for which f(X) = 7.  The trick is to interpose a 
                   function between ZeroIn() and Blackbox() that will 
                   return a 0 to ZeroIn() whenever Blackbox() returns 
                   7. By calling that function, ZeroIn() finds a 
                   value of X for which Blackbox( X ) = 7, as 
                   required:
                     Result = ZeroIn("Temp", ;
                                       <other parameters omitted>)
                 
                     FUNCTION Temp( nQ )
                     RETURN Blackbox( nQ ) - 7
     Written for.: dBASE IV Version 1.5
     Rev. History: Original function 1990.
                   Modified to take optional parameters, 4/13/1992
                   Fall, 1994 -- revised for dBASE 5.0 for Windows.
                                 Jay Parsons, 72662,1302 
                   11/09/1996 -- Modified to work with this class.
     Calls.......: The function whose name is first parameter.
     Usage.......: math.ZeroIn( <cFunction>, <fX1>, <fX2>, ;
                                <fAbserror>, <nMaxiter>, <n_Flag> ;
                                [, xPass1 [, xPass2 [, xPass3 ] ] ] )
     Example.....: ? _app.Math.ZeroIn( "Npv", 0, 200, .000001, 200,;
                                     n_Flag, 11 )
     Returns.....: a float value representing a root, if n_Flag < 3.
     Parameters..: cFunction = the name of the function to solve for 
                               a root.
                   fX1       = one of the X-values between which the
                               root is sought.
                   fX2       = the second of these values.
                   Note: These MUST be chosen so the f( X ) values 
                   for the two of them have opposite signs (they 
                   must bracket the result).
                   fAbserror  = the absolute error allowed in the 
                                result.
                   nMaxiter   = the maximum number of times to 
                                iterate.
                   n_Flag     = an integer to signal success ( < 3 ) 
                                or failure.
                   xPass1 . . . 3 = arguments to be passed through to
                                    cFunction.
                   The parameter "n_Flag" should be passed as a 
                   variable so it may be accessed on return.  The 
                   limit of 9 literal parameters may require passing 
                   others as variables.  The "xPass" parameters are 
                   optional and the fact there are three of them is 
                   arbitrary; they exist to hold whatever parameters
                   may be needed by the function cFunction being 
                   called aside from the value of X for which it is 
                   being evaluated. Add more and change the 3 
                   "&cFunc." lines below if you need more.
     Side effects: Uses and alters a global numeric variable, here 
                   called "n_Flag", to report error conditions 
                   resulting in value returned being meaningless.  
                   Possible n_Flag values are:
                         1   success - root found within error 
                                       allowed
                         2   success - root was found exactly
                         3   error   - function value not converging
                         4   error   - original values do not bracket
                                       a root
                         5   error   - maximum iterations exceeded
    ----------------------------------------------------------------- */
      
      local nSplits, fBracket, fFary, fNeary, nIters
      local fMaxabs, fOldx, fOldy, fDiffx, fAbsdiff, fSecant
      
      store 0 to nSplits, nIters
      fBracket = abs ( fNearX - fFarX )
      fFarY = &cFunc.( fFarX, xPass1, xPass2, xPass3 )
      fNearY = &cFunc.( fNearX, xPass1, xPass2, xPass3 )
      
      if sign( fNearY ) = sign( fFarY )
         n_Flag = 4
         RETURN float(0)
      endif
      
      fMaxAbs = this.max( this.abs( fNearY ), this.abs( fFarY ) )
      n_Flag = 0
      
      // Main iteration loop
      do while true
         
         if this.abs( fFarY ) < this.abs( fNearY )
             
            /* Interchange m->fNearX and fFarX so that
               m->fNearX is closer to a solution--
               abs( fNearY ) <= abs( fFarY ) */
            
            fOldX  = fNearX
            fOldY  = fNearY
            fNearX = fFarX
            fNearY = fFarY
            fFarX  = fOldX
            fFarY  = fOldY
         endif
         
         fDiffX = fFarX - fNearX
         fAbsDiff = this.abs( fDiffX )
         
         // Test whether interval is too small to continue
         
         if fAbsDiff <= 2 * fAbsErr
            if this.abs( fNearY ) > fMaxAbs
               
                // Yes, but we are out of bounds
                
                n_Flag = 3
                fNearX = float(0)
             else
                
                // Yes, and we have a solution!
                
                n_Flag = 1
             endif
             exit
          endif
          
          // Save the last approximation to x and y
          
          fOldX = fNearX
          fOldY = fNearY
          
          /* Check if reduction in the size of
             bracketing interval is satisfactory.
             if not, bisect until it is. */
          
          nSplits = nSplits + 1
          if nSplits >= 4
             if 4 * fAbsDiff >= fBracket
                fNearX = fNearX + fDiffX / 2
             else
                nSplits = 0
                fBracket = fAbsDiff / 2
                
                // Calculate secant
                
                fSecant = ( fNearX - fFarX ) * fNearY ;
                                    / ( fFarY - fNearY )
                
                // But not less than error allowed
                
                if this.abs( fSecant ) < fAbsErr
                   fNearX = fNearX + fAbsErr * sign( fDiffX )
                else
                   fNearX = fNearX + fSecant
                endif
             endif
          endif
          
          // Evaluate the function at the new approximation
          
          fNearY = &cFunc.( fNearX, xPass1, xPass2, xPass3 )
           
          // if it's exactly zero, we win!  Run with it
          
          if fNearY = 0.00
             n_Flag = 2
             exit
          endif
          
          /* else adjust iteration count and quit if too
             many iterations with no solution */
           
          nIters = nIters + 1
          if nIters > nMaxIter
             n_Flag = 5
             fNearX = float( 0 )
             exit
          endif
          
          /* And finally keep as the new m->fFarX that one
             of the previous approximations, m->fFarX and
             fOldx, at which the function has a sign opposite
             to that at the new approximation, fNearx. */
          
          if sign( fNearY ) = sign( fFarY )
             fFarX = fOldX
             fFarY = fOldY
          endif
       enddo
       
   RETURN fNearX
   // -- EoM: Zeroin()

   /*
      Programmer..: Michael Nuwer
      Date........: September, 2004
      Notes.......: Future Value function, works like the one
                    built in to Excel.
      Written for.: dBASE Plus
      Rev. History: September, 2004 -- Original
      Calls.......: None
      Usage.......: math.ExcelFV( nRate, nPeriods, nPMT, nPV, nType )
      Example.....: Assume an annual interest rate of 8%; a 
                    payment of $100 per month; an opening balance 
                    of $20; and the payment is make at the 
                    beginning of the month
                    nVal = _app.Math.ExcelFV( .08/12,12,100,20,1)
      Returns.....: Future Value
      Parameters..: nRate    -- the interest rate per period
                    nPeriods -- the total number of payment periods
                                in an annuity
                    nPMT     -- the unchanging payment made each 
                                period
                    nPV      -- the present value, or the opening 
                                balance of an annuity.   
                    nType    -- a value representing the timing of 
                                the payment. 
                                1              = payment is made 
                                                 at the beginning 
                                                 of the period; 
                                0 (or omitted) = payment is 
                                                 made at the end 
                                                 of the period.
   */

   function ExcelFV( nRate, nPeriods, nPMT,  nPV, nType )
      local nValue

      nPMT  = iif(nPMT=NULL,0,nPMT)
      nType = iif(nType=NULL,0,nType)
      nPV   = iif(nPV = NULL,0,nPV)

      nValue = 0
      for i=1 to nPeriods-1
          nValue += ((nPMT) * ( 1 + (nRate)) ^ (i))
      endfor

      if nType = 1
         nValue += (nPMT) * ( 1 + (nRate))^nPeriods
      else
         nValue += nPMT
      endif

      if nPV > 0
         nValue += nPV * ( (1+nRate)^nPeriods )
      endif

   return nValue * -1
   // EoM: ExcelFV()

   /*
      Programmer..: Michael Nuwer
      Date........: September, 2004
      Notes.......: Present Value function, works like the one
                    built in to Excel.
      Written for.: dBASE Plus
      Rev. History: September, 2004 -- Original
      Calls.......: None
      Usage.......: math.ExcelPV( nRate, nPeriods, nPMT, nPV, nType )
      Example.....: Assume an annual interest rate of 8%; a 
                    payment of $100 per month; an opening balance 
                    of $20; and the payment is make at the 
                    beginning of the month
                    nVal = _app.Math.ExcelPV( .08/12,12,100,0,0)
      Returns.....: Present Value
      Parameters..: nRate    -- the interest rate per period
                    nPeriods -- the total number of payment periods
                                in an annuity
                    nPMT     -- the unchanging payment made each 
                                period
                    nPV      -- the present value, or the opening 
                                balance of an annuity.   
                    nType    -- a value representing the timing of 
                                the payment. 
                                1              = payment is made 
                                                 at the beginning 
                                                 of the period; 
                                0 (or omitted) = payment is 
                                                 made at the end 
                                                 of the period.
    */


   function ExcelPV( nRate, nPeriods, nPMT,  nFV, nType )
      local nValue, n

      nPMT  = iif(nPMT=NULL,0,nPMT)
      nType = iif(nType=NULL,0,nType)
      nFV   = iif(nFV = NULL,0,nFV)

      nValue = 0
      if nType = 1
         n = nPeriods - 1
         // last payment; made at end of period
         // no interest accumulation
         nValue = nPMT
      else
         n = nPeriods
      endif   

      for i=1 to n
          nValue += nPMT / ( (1+nRate)^i )
      endfor

      if nFV > 0
         nValue += nFV / ( (1+nRate)^nPeriods )
      endif

   return nValue * -1
   // EoM: ExcelPV()


ENDCLASS
// -- End of Class: MathEx


/* ---------------------------------------------------------------------
   End of File: MATHEX.CC
   --------------------------------------------------------------------- */