// Metric: 0 - Chars or 6 - Pixels only for TopicInfoButton and
//                                          kmCallDiacrit,
//         Other buttons are non-specific. The InfoForm control
//         is a separate dialog and does not need to have 
//         the metrics messed with.

// -- File: MOREBTNS.CC

/* ----------------------------------------------------------------------
    Description:
 
       This UI file is designed to give some optional buttons
       which can be used either 'as is' or you can even subclass them.
 
    Programmers: 
 
       Ken Mayer
       Jose Luis Perez
 
    History: 
 
       11/28/1996 -- First version of this custom class file.
       02/07/1997 -- Jose (above) found some problems in
                     the Copy2Clip button, and sent me the
                     fix ...
       03/24/1998 -- Added kmCallDiacrit
 
    Instantiation: 
 
       set procedure to morebtns.cc additive
       * create/modify a form, place an instance of one of
       * these buttons on the form as a normal button ...
 
    Classes:
       More details on each of these is contained in the commentary for
       each class ...
 
       Copy2ClipButton
         Copies contents of value property for every object on form
         into the Windows clipboard, which can then be pasted into
         another software package (such as a Word document ...).
 
       TopicInfoButton
         Modification of Borland's Info Button in the BUTTONS.CC
         file, this one is more flexible.

       kmCallDiacrit
         Button that calls DIACRIT.WFM (a separate form), and
         sets the value returned from that form into the 
         "current" control, if there is one ...
   -------------------------------------------------------------------- */


CLASS Copy2ClipButton( oParent ) OF PUSHBUTTON( oParent ) CUSTOM
/* 
   ----------------------------------------------------------------------
    Description:
 
       This button will attempt to place the value property of every 
       standard control on your form (that has a 'value' property)
       into the clipboard of Windows, with a line feed (chr(13)) to
       break each item out.
 
       This code is based very loosely on a message in the Visual dBASE
       newsgroup on the Internet: comp.databases.visual-dbase; the
       messages included have authors of: Edward Cross, Alan Katz,
       and Ned Buyukmihci(?).
 
      NOTE: This button may not work with containers or notebooks --
             if using a notebook, you may want to set thisPageOnly
             to false -- otherwise you will only get what's on
             the same page of the notebook that the form is currently
             on (i.e., if you have a two page form, with a notebook
             on page 1, and you have items on several pages of
             the notebook, you would probably only get the items
             on the first page of the notebook ... similarly,
             if the notebook is on page 2, you would only get
             the items on the second page of the notebook ...).
    
 -- Programmer:
 
       Ken Mayer (with thanks to Jim Sare for reminding of the 
                  simple things ... >duh<)
       Jose Luis Perez -- fixes ... (see below)
 
 -- History:
 
       11/25/1996 -- Original
       01/01/1997 -- Minor modifications (thanks to Jim Sare again)
                     to make it work on Modal forms ...
       02/06/1997 -- (Jose Luis Perez) Changes  all references from
                     CopyIt1 to CopyIt. Changes in the OnOpen event.
                     Simplified logic in value selection.
       01/07/1998 -- Minor mods for VdBASE 7
 
 -- Methods:
 
 --    Init()
         Initialize entryfield needed.
 
 --    CopytoClipBoard()
         Copies the items on the form into the clipboard, this uses
         a version of Bowen Moursund's "enabled" routine to loop 
         through the objects on the form, looking for ones that have
         a value property. It will create an entryfield object that
         is disabled and invisible (the user can't see it or interact
         with it).
 
 -- Properties:
 
 --    ThisPageOnly
         Logical, defaults to true -- only copy the contents of the
         current page. If set to false (.f.), it will copy the contents
         of the value property of all objects on the form to the
         clipboard (this may be more information than you wanted ...).
         See notes above -- if used on a Notebook, you may need/want
         to set this property to False.
 ----------------------------------------------------------------------
*/
   protect CopytoClipBoard
   this.ThisPageOnly   = True
   this.Height         = 1.54
   this.Width          = 4
   this.UpBitMap       = "RESOURCE #2040"
   this.DownBitMap     = "RESOURCE #2041"
   this.DisabledBitMap = "RESOURCE #2041"
   this.Text           = ""
   this.SpeedTip       = "Copy items on form to clipboard"
   this.onClick        = class::CopytoClipBoard
   #if __version__ = 2.6
      // for these toolbuttons, don't use the system theme ...
      this.systemTheme := false // for dBASE Plus 2.6 
   #endif
  
   FUNCTION OnOpen
      *-- initialize 'CopyIt' entryfield, make it
      *-- invisible, so user doesn't
      *-- even know it exists ...
      if type( "form.CopyIt" ) # "O"
         define EntryField CopyIt of Form ;
            property;
               visible .f.,;
               enabled .t.
      endif

   RETURN NULL

   FUNCTION CopytoClipBoard
      form.CopyIt.value = ""  && empty it out 

      private oControl
      *-- get first control
      oControl = form.first

      *-- loop through objects on form:
      do
         *-- if value property is undefined, never mind
         *-- don't process if name of control contains 'copyit'
         if ( type( "oControl.value" ) # "U" ) and ;
            ( upper(oControl.Name) # "COPYIT" ) 

            if ( .not. this.ThisPageOnly) .or. ;
               ( oControl.PageNo == form.PageNo )
               *-- check for types and convert if necessary to string
               do case
                  case type( "oControl.value" ) $ "NF" && numeric/float
                       cValue = ltrim(str( oControl.value ))
                  case type( "oControl.value" ) = "L"  && logical
                       cValue = iif( oControl.value, "True", "False" )
                  case type( "oControl.value" ) = "D"  && date
                       cValue = dtoc( oControl.value )
                  otherwise                            && char/memo
                       cValue = oControl.Value
               endcase
               // Add control name ...
               if type( "oControl.Text" ) # "U" and ;
                  not empty( oControl.Text )
                  cValue = trim( oControl.Text ) + ": " + cValue
               else
                  cValue = oControl.Name + ": " + cValue
               endif
               *-- add to end of whatever is currently in entryfield                                 
               if not empty( cValue )
                  form.CopyIt.value = form.CopyIt.value + ;
                                      cValue + chr(13)+chr(10)
               endif // not empty
            endif && page check
         endif && type ...
         *-- next control
         oControl = oControl.before

         *-- end of loop
      until oControl.Name == form.first.name
      
      *-- "highlight" everything in entryfield:
      form.CopyIt.keyboard( "{HOME}{SHIFT-END}" )

      *-- now we have it all, let's put it in the clipboard:
      form.CopyIt.copy()

      *-- let user know:
      msgbox( "Record is copied to buffer -- you can paste it into "+;
              "another Windows application now (Ctrl+V).", "Copied", 48)

   RETURN NULL

ENDCLASS
//-- End of Class: COPY2CLIPBUTTON


CLASS TopicInfoButton(oParent) OF PUSHBUTTON(oParent) Custom
*----------------------------------------------------------------------
*-- Description:
*
*      Button for displaying information about the currently running 
*      form. This button instantiates an InfoForm object, defined below
*
*-- Programmer:
*
*      Borland Samples Group
*      Ken Mayer
*
*-- History:
*
*      ?? -- Shipping version in \VISUALDB\SAMPLES (VdBASE 5.5) directory
*      05/??/1996 -- Modified for Borland Developer's Conference
*                    with permission of Borland Samples Group
*      12/01/1996 -- minor (very) modifications for
*                    the dUFLP library (just layout ...)
*
*-- Properties:
*     Both of these are required, and should be instantiated in either
*     the button or form's OnOpen:
*
*--   InfoFile
*       Text file that contains information needed
*
*--   InfoName
*       Topic to display on info form ...
*
*
*-- Example:
*
*      PROCEDURE infobutton1_onopen
*           This.TopicName = "Topic 1"
*           This.TopicFile = "MyHelpFile.txt"

*-- NOTES:
*      In the files used to store the information, this set of 
*      routines assumes a layout of:
*
*       TOPIC 1
*       -------
*       The line above is assumed to be there, if it isn't, the
*       first line of this topic would not be displayed.
*       Each topic must end with a specific character, which
*       defaults to a star (*). This must be on its own line at
*       the beginning of the line.
*       *
*
*       TOPIC 2
*       -------
*       A second topic that can be displayed from another instance
*       of the info button.
*       *
*
*     If you wish to change the button from the 'info' button to
*     a help button, you can either uncomment the statement in the
*     constructor code below, or in the form's (or button's) onOpen
*     method, set the bitmap:
*        form.infobutton1.UpBitMap = "resource #32"
*
*-- Dependencies:
*
*--   INFOFORM -- included here in 'MOREBTNS.CC'
*----------------------------------------------------------------------

   this.Height   = 1.2
   this.Width    = 3.5
   this.UpBitmap = "Resource #112" && 'i' (info)
   *this.UpBitmap = "Resource #32" && question mark
   this.Text     = ""
   this.SpeedBar = .T.  && can't receive "focus"
   this.SpeedTip = "Information About This Form/Topic"
   this.PageNo   = 0

   * Custom Properties
   this.TopicName = ""
   this.TopicFile  = ""

   #if __version__ = 2.6
      // for these toolbuttons, don't use the system theme ...
      this.systemTheme := false // for dBASE Plus 2.6 or later
   #endif

   function onDesignOpen( lFromPalette )
      if lFromPalette and form.metric == 6 // Pixels
         this.height := 26
         this.width  := 25
      endif
   return

   *********************************************************************
   PROCEDURE OnClick
   *********************************************************************
   local f

   f = new InfoForm(this.TopicName,this.TopicFile)
   if .not. empty(f.fileName)   && Information about the topic exists
      f.Top = -300              && open off screen so it will snap
                                && onscreen using Jim's CENTERFORM()
      f.ReadModal()
   endif

ENDCLASS

***********************************************************************
CLASS INFOFORM(TopicName,TopicFile) OF FORM
// Metrics: As this is a separate dialog form, there is no need
// to attempt to deal with the metrics for this -- it works "as is",
// no need ot change it.
   * Class for displaying information.  This information is extracted 
   * from File: "TopicFile" (parameter passed to this routine).
   ********************************************************************
   this.autoCenter  = true
   this.Height      = 7.2344
   this.ColorNormal = "GB+"
   this.PageNo      = 1
   this.Text        = TopicName
   this.TopMost     = .F.
   this.Width       = 65
   this.Left        = 18
   this.MDI         = .F.
   this.Sizeable    = .F.
   this.Maximize    = .F.
   this.Minimize    = .F.
   this.Sysmenu     = .F.
   this.OnClose     = CLASS::Form_OnClose
   this.OnOpen      = CLASS::Form_OnOpen
   this.TopicFile   = TopicFile

   DEFINE EDITOR INFOEDITOR OF THIS;
       PROPERTY;
         CUATab .T.,;
         Top 1.2,;
         Height 6.0,;
         FontBold .F.,;
         ColorNormal "B+/GB+",;
         PageNo 1,;
         Border .F.,;
         Width 64,;
         Wrap .F.,;
         Left 1,;
         Value "",;
         Modify .F.,;
         FontSize 10,;
         FontName "Arial"

   DEFINE PUSHBUTTON CLOSEBUTTON OF THIS;
       PROPERTY;
         Top 0,;
         Height 1.1172,;
         ColorNormal "BtnText/BtnFace",;
         PageNo 1,;
         Text "",;
         Width 3.5,;
         UpBitmap "RESOURCE #36",;
         Group .T.,;
         Left 0,;
         OnClick {;form.Close()}

   *--------------------------------------------------------------------
   *-- The method "CreateTempFile" is the meat of this class --
   *-- it is the code that extracts the topic information ...
   *--------------------------------------------------------------------
   this.fileName = CLASS::CreateTempFile(TopicName,TopicFile)
   if .not. empty(this.fileName)
      this.infoEditor.dataLink = "file " + this.fileName
   endif

   *********************************************************************
   PROCEDURE Form_OnOpen

     *-- set defaults
     form.infoeditor.height = form.height - 1.22
     form.infoeditor.width  = form.width  - 1

   *********************************************************************
   PROCEDURE Form_OnClose

   * Close info form, and delete temp file, if it exists.
   *********************************************************************
   private cFileName

   if .not. empty(this.fileName)
      cFileName = this.fileName
      this.infoEditor.release()
      erase &cFileName
   endif  && .not. empty(this.filename)

   *********************************************************************
   FUNCTION CreateTempFile(TopicName,TextName)
      
   * Create temporary file containing information about TopicName, 
   * extracted from "TextName". This temporary file is then displayed
   * in the editor object on the form.
   *********************************************************************
      
      *----------------------------------------------------------------
      *-- We can change the end of topic character from an asterisk --
      *-- by changing it here. This should be the only place this would  
      *-- be necessary. If, for example, your topic needed to have some
      *-- asterisks at the beginning of a line, we could change the
      *-- end of topic character to, say, a carat (^) or something.
      *----------------------------------------------------------------
      #define END_TOPIC  "*"
      private cTempFileName, fhTopicTxt, fhTempFile, cLine 
      private lTopicFound, result
      
      *----------------------------------------------------------------
      *-- The code below uses low-level file functions to open the
      *-- text file, and find the information for the topic,
      *-- creating a temporary file to save the information into.
      *-- The temporary file is what is actually used for the 
      *-- datasource for the editor.
      *----------------------------------------------------------------
      cTempFileName = ""   && init the filename memvar.
      
      *----------------------------------------------------------------
      *-- assign a file-handle to the file if we have a successful
      *-- open. If the value returned is zero, we have an error.
      *----------------------------------------------------------------
      if .not. file(textname)
         msgbox(TextName+" was not found.","Info",48)
         RETURN ""
      endif
      fhTopicTxt = fopen(TextName, "R")  && Open file Read-only
      if fhTopicTxt = 0
         msgbox(TextName+" was not found, or couldn't be "+;
                            "opened.","Info",48)
      else
         *-------------------------------------------------------------
         *-- since we didn't have an error, we scan the file
         *-- searching line-by-line until we find the topic heading.
         *-------------------------------------------------------------
         lTopicFound = .F.
         do while .not. lTopicFound .and. .not. feof(fhTopicTxt)
            *-- Read each line (fGets)
            cLine = fgets(fhTopicTxt)    && Search until find TopicName
            *-- if we found the topic ...
            if upper(rtrim(cLine)) = upper(TopicName)
               lTopicFound = .T.
            endif
         enddo
         *-------------------------------------------------------------
         *-- if the memvar lTopicFound still contains the value .f.,
         *-- we did not find the topic we are looking for. This means
         *-- we should display an error message ...
         *-------------------------------------------------------------
         if .not. lTopicFound                && Reached end of file
            msgbox("Information about " + TopicName +;
                   " was not found in "+TextName+".",;
                   "Sorry",48)
         else                                && Save info to temp file
            *----------------------------------------------------------
            *-- If we're here, let's create the temporary file, and
            *-- store the topic information to that file, reading line-
            *-- by-line in the original, and writing that information 
            *-- out to the temporary file.
            *----------------------------------------------------------
            cTempFileName = fUnique()  && get a unique temp filename
                                       && spiffy function built-in to 
                                       && VdB
            fhTempFile = fCreate(cTempFileName, "RW")
            if fhTempFile = 0                 && Couldn't create file
               *-------------------------------------------------------
               *-- This usually only occurs if the file by that name
               *-- exists, or there's a RIGHTS issue on a LAN -- on
               *-- a LAN for this to work, you need to have create and
               *-- erase rights ...
               *-------------------------------------------------------
               msgbox("Couldn't create temp file.",;
                      "Sorry",48)
            else                             && Temp file was created
               *-------------------------------------------------------
               *-- This is where we start processing. Now, the original
               *-- layout uses an underscore line (actually a set of
               *-- dashes) to separate the topic from the text. This
               *-- first cLine = ... is a throw-away -- we don't wish
               *-- to store the underscore line in the output file.
               *-- NOTE: If you do not wish to use an underscore for
               *-- your topic heading, comment the following line out!
               *-------------------------------------------------------
               cLine = fgets(fhTopicTxt)     && Underscore Line
               *-------------------------------------------------------
               *-- This is the first line of the actual topic in the
               *-- text file. 
               *-------------------------------------------------------
               cLine = fgets(fhTopicTxt)
               *-------------------------------------------------------
               *-- Loop until we find the end of topic character, 
               *-- and/or we hit the end of file ...
               *-------------------------------------------------------
               do while rtrim(cLine) <> END_TOPIC .and.;
                        .not. feof(fhTopicTxt)
                  *----------------------------------------------------
                  *-- Store the current line in the output/temp file
                  *----------------------------------------------------
                  result = fputs(fhTempFile, cLine)
                  *----------------------------------------------------
                  *-- Get the next line and then loop ...
                  *----------------------------------------------------
                  cLine = fgets(fhTopicTxt)
               enddo  && while rtrim(cLine) ...
               
               *-------------------------------------------------------
               *-- Close the temp file. We're done with it now.
               *-------------------------------------------------------
               result = fclose(fhTempFile)
            endif  && fhTempFile = 0
         endif  && .not. lTopicFound
         *-------------------------------------------------------------
         *-- Close the topic file
         *-------------------------------------------------------------
         result = fclose(fhTopicTxt)
      endif  && fhTopicTxt = 0
      
   RETURN cTempFileName  && return the name of the temp file we
                         && just created ...

ENDCLASS
*-- End of Classes: INFOBUTTON/INFOFORM

class kmCallDiacrit(parentObj) of PUSHBUTTON(parentObj) custom
/*
    kmCallDiacrit
    This custom button can be placed on
    a form, and then used to insert a character from the
    upper ANSI character set. 

    Updated: January 27, 2002, to work around a bug with
    MDI forms losing focus completely when calling 
    a modal dialog form. This workaround is simply
    to store the form's activeControl as a reference,
    and set focus when the diacrit form closes.

    Caution: There is no easy way to encapsulate this 
    so that it can check the current entryfield/control
    on a form to see if it can accept characters -- if
    it can't, then the error will be something you will
    have to deal with. There is a trap to see if we're
    in edit or append mode for the form's rowset, but if
    the last control the user touched was, say, a listbox,
    then trying to keyboard characters to it may be problematic.
    One way around this might be to add some code, or copy this
    and tinker with it, perhaps add a form property
    that gets set in a control's onFocus or onLostFocus
    to enable/disable the button ...

    *********************
    REQUIRES: DIACRIT.WFM
    *********************

    Side-effects: creates custom property of _app called
    cDiacrit (e.g., _app.cDiacrit ). The reason it's
    a property of _app, instead of form is that we 
    are really using two forms and this is the easiest
    way to get them to both be able to see this
    property. You may want to stub out this property
    when you close the parent form with:  
    _app.cDiacrit = null

*/

   with (this)
      onClick = class::CALLDIACRIT_ONCLICK
      height = 1.0909
      width = 10.4286
      text = "Diacriticals"
      speedTip = "Insert upper-ANSI characters to current control"
      metric = 0	// Chars
      speedBar = true
      value = false
   endwith
   #if __version__ = 2.6
      this.systemTheme := false
   #endif

   function onDesignOpen( lFromPalette )
      if lFromPalette and form.metric == 6 // Pixels
         this.height := 24
         this.width  := 73
      endif
   return


   function CALLDIACRIT_onClick

      // Are we editing or appending? If not, don't allow this
      // (NOTE: this test is valid only if we have a rowset on
      // the form, otherwise it doesn't get checked):
      if type( "form.rowset" ) = "O" and ;
        ( form.rowset.state < 2 or form.rowset.state > 3 ) 
         msgbox( "You are not entering or editing data.",;
                 "Can't do this!", 16)
         return
      endif

      // clear this out if it exists, otherwise initialize it:
      _app.cDiacrit = ""

      // run form (modal):
      do :duflp:diacrit.wfm with true

      // if we returned a value:
      if not empty( _app.cDiacrit )
         keyboard "{HOME}"       // beginning of field
         keyboard _app.cDiacrit  // keyboard character
         // -----------------------------------------------
         // With the character at the beginning of the
         // field, it's obvious, and can then be copied/
         // pasted to the correct location.
         // -----------------------------------------------

         // if you don't want the dialog to appear, 
         // comment out this statement:
         msgbox( "Character '"+_app.cDiacrit+"'is at beginning of "+;
                 "field -- use cut/paste to place it where you really "+;
                 "need it.","Done!", 48 )
      endif

      // set focus (due to bug in dB2K):
      form.activeControl.setFocus()

   return
endclass
// End of class: kmCallDiacrit

class cursor_on_button(f) of pushbutton(f)
/*  Description:
       This button will move the cursor to its center
       when the form opens

    Programmers: 
           
       Romain Strieff
 
    History: 
 
       09/07/1999 -- First version

    Instantiation: 
 
       set procedure to Curson_On_Button.cc additive
       create/modify a form, place an instance of 
       this button on the form

     Remark:
        use only _1_ instance on this button on any 
        form. If you use several, obviously only the
        last one in the Z-order will have the cursor.       

*/

   if type("GetWindowRect") # "FP"
      extern cLOGICAL GetWindowRect;
      (cHandle, cptr) user32
   endif
   if type("SetCursorPos") # "FP"
      extern cLOGICAL SetCursorPos;
      (CINT, CINT) user32
   endif
   if type("GetClientRect") # "FP"
      extern cLogical GetClientRect;
      (cHandle, cPtr) user32
   endif
   #define GCR_WIDTH    5
   #define GCR_HEIGHT   7
   #ifndef	MAKE8
      #define	make8(cBuff, nIndex)	;
      asc(substr(cBuff, nIndex))
   #endif
   #ifndef	MAKEU16
      #define	makeu16(cBuff, nIndex);
      make8(cBuff, nIndex) + bitlshift(make8(cBuff,;
      nIndex + 1), 8)
   #endif

   Function Onopen
      local cRect,nCol,nRow,nPixWidth,nPixHeight
      //create receiving buffer
      cRect=space(8)
      //Get the value of the button rect values
      GetClientRect(this.hwnd, cRect)  
      //calculate width and height
      nPixWidth  = makeu16(cRect, GCR_WIDTH)  - 1
      nPixHeight = makeu16(cRect, GCR_HEIGHT) - 1
      //Get the position of the button onscreen
      GetWindowRect(this.hwnd,cRect)
      //extract it from the Unicode value
         nCol = cRect.GetByte(0) +;
              BitLShift(cRect.GetByte(1), 8) +;
              BitLShift(cRect.GetByte(2), 16) +;
              BitLShift(cRect.GetByte(3), 24)
         nRow = cRect.GetByte(4) +;
              BitLShift(cRect.GetByte(5), 8) +;
              BitLShift(cRect.GetByte(6), 16) +;
              BitLShift(cRect.GetByte(7), 24)

      //position the mousepointer in the middle
      SetCursorPos(ncol+(nPixWidth/2),;
                   nrow+(nPixHeight/2))   
   return

endclass
// End of class: cursor_on_button

*----------------------------------------------------------------------
*-- End of File: MOREBTNS.CC
*----------------------------------------------------------------------
